/**
 * y25 - Visualizing Graphs in 2.5 Dimensions
 * 
 * Arbeitsbereich: Paralleles Rechnen Prof. Kaufmann
 * Studienarbeit:  Werner Jainek
 * Betreuung:      Nina Lehmann
 *
 */
package y25.base;

import y.base.DataProvider;
import y.base.Edge;
import y.base.EdgeCursor;
import y.base.EdgeMap;
import y.base.Graph;
import y.base.Node;
import y.base.NodeMap;
import y.view.EdgeRealizer;
import y.view.Graph2D;
import y.view.NodeRealizer;

/**
 * A utility class which provides means to duplicate entire <code>Graph</code>s
 * or <code>Graph2D</code>s. This is necessary, because the
 * <code>clone</code> methods provided by those classes do not take into
 * account <code>DataProvider</code>s that are associated to the respective
 * graphs and therefore remain uncopied.
 * <p>
 * To copy a Graph, you call {@link #DuplicateGraph DuplicateGraph},
 * {@link #DuplicateGraph2D DuplicateGraph2D} or
 * {@link #ConvertGraphToGraph2D ConvertGraphToGraph2D} and specify the
 * <code>DataProvider</code>s that should be copied. This is accomplished via
 * the {@link GraphDataInfo} class.
 * 
 * @author Werner Jainek
 * 
 */
public class GraphTools {

	static private void CopyKeys(Graph graph, Graph copy,
			GraphDataInfo[] dataInfo) {

		// did we get any data to copy at all ?
		if (dataInfo != null) {

			// get arrays for easy indexing
			Node[] nodeArray = graph.getNodeArray();
			Node[] copyNodeArray = copy.getNodeArray();
			Edge[] edgeArray = graph.getEdgeArray();
			Edge[] copyEdgeArray = copy.getEdgeArray();

			// copy every requested DataProvider by creating a new Node/Edgemap
			// in the new graph
			for (int i = 0; i < dataInfo.length; ++i) {
				// get current dataprovider
				DataProvider p = graph.getDataProvider(dataInfo[i].key);

				// determine if we want to copy node or edge values
				switch (dataInfo[i].mapType) {

				case GraphDataInfo.NODEMAP:
					// copy values for all nodes
					NodeMap nm = copy.createNodeMap();
					for (int j = 0; j < copy.N(); ++j) {
						Node source = nodeArray[j];
						Node target = copyNodeArray[j];
						// determine the type of the value to copy for each node
						switch (dataInfo[i].valueType) {
						case GraphDataInfo.BOOLEAN:
							nm.setBool(target, p.getBool(source));
							break;
						case GraphDataInfo.DOUBLE:
							nm.setDouble(target, p.getDouble(source));
							break;
						case GraphDataInfo.INTEGER:
							nm.setInt(target, p.getInt(source));
							break;
						}
					}
					// add new nodemap as dataprovider
					copy.addDataProvider(dataInfo[i].key, nm);

					break;

				case GraphDataInfo.EDGEMAP:
					// copy values for all edges
					EdgeMap em = copy.createEdgeMap();
					for (int j = 0; j < copy.E(); ++j) {
						Edge source = edgeArray[j];
						Edge target = copyEdgeArray[j];

						// determine the type of the value to copy for each edge
						switch (dataInfo[i].valueType) {
						case GraphDataInfo.BOOLEAN:
							em.setBool(target, p.getBool(source));
							break;
						case GraphDataInfo.DOUBLE:
							em.setDouble(target, p.getDouble(source));
							break;
						case GraphDataInfo.INTEGER:
							em.setInt(target, p.getInt(source));
							break;
						}
					}
					// add new edgemap as dataprovider
					copy.addDataProvider(dataInfo[i].key, em);
					break;

				} // switch

			} // for

		} // if keys!=null

	}

	// Important: only supply empty graphs as targets
	static private void CopyGraphToGraph(Graph graph, Graph copy,
			GraphDataInfo[] dataInfo) {

		// check if there is anything to do
		if (graph == null || copy == null)
			return;

		// --- COPY ALL NODES ---
		for (int i = 0; i < graph.N(); ++i) {
			copy.createNode();
		}

		// --- COPY ALL EDGES ---

		// get node array of copygraph for easy indexing
		Node[] copyNodeArray = copy.getNodeArray();

		// copy all edges
		for (EdgeCursor ec = graph.edges(); ec.ok(); ec.next()) {
			Edge e = ec.edge();
			Node copySourceNode = copyNodeArray[e.source().index()];
			Node copyTargetNode = copyNodeArray[e.target().index()];
			copy.createEdge(copySourceNode, copyTargetNode);
		}

		// sort edges to be in correct order
		// TODO !!!

		// --- COPY ALL KEYS ---
		CopyKeys(graph, copy, dataInfo);

	}

	static private void CopyGraphToGraph2D(Graph graph, Graph2D graph2d,
			GraphDataInfo[] dataInfo) {

		// check if there is anything to do
		if (graph == null || graph2d == null)
			return;

		// this method is only here in case there is something else to do when
		// copiying graphs to 2d, other than assigning node/edge-realizers,
		// which is done automatically by using the default node/edge-realizers
		// of the empty graph2d
		CopyGraphToGraph(graph, graph2d, dataInfo);
	}

	// Important: only supply emty graphs as targets
	static private void CopyGraph2DToGraph2D(Graph2D graph, Graph2D copy,
			RealizerFactory realizerFactory, GraphDataInfo[] dataInfo) {

		// check if there is anything to do
		if (graph == null || copy == null)
			return;

		// --- COPY ALL NODES ---
		Node[] nodeArray = graph.getNodeArray();
		for (int i = 0; i < nodeArray.length; ++i) {

			// create a copy of current node realizer (possibly a child class is
			// instanciated by the factory, e.g. a 25D version)
			NodeRealizer nr = graph.getRealizer(nodeArray[i]);
			NodeRealizer nrcopy = realizerFactory.copyNodeRealizer(nr);
			copy.createNode(nrcopy);
		}

		// --- COPY ALL EDGES ---

		Edge[] edgeArray = graph.getEdgeArray();
		Node[] copyNodeArray = copy.getNodeArray();
		for (int i = 0; i < edgeArray.length; ++i) {
			// determine source and target nodes
			Edge e = edgeArray[i];
			Node copySourceNode = copyNodeArray[e.source().index()];
			Node copyTargetNode = copyNodeArray[e.target().index()];

			// create a copy of current edge realizer (possibly a child class is
			// instanciated by the factory, e.g. a 25D version)
			EdgeRealizer er = graph.getRealizer(edgeArray[i]);
			EdgeRealizer ercopy = realizerFactory.copyEdgeRealizer(er);
			copy.createEdge(copySourceNode, copyTargetNode, ercopy);
		}

		// sort edges to be in correct order
		// TODO !!!

		// --- COPY ALL KEYS ---
		CopyKeys(graph, copy, dataInfo);
	}

	/**
	 * Creates a copy of the given Graph.
	 * <p>
	 * All nodes and edges of the given graph are copied. The index of a both
	 * nodes and edges is the same in the duplicate graph. The in/out order of
	 * edges is kept as well (! This is not true yet !).
	 * <p>
	 * Furthermore, the user can specify which DataProviders should be copied by
	 * providing their respective keys. For each data provider, the user also
	 * has to specify if it assigns values to nodes or to edges and the type of
	 * the values. This is done via the {@link GraphDataInfo GraphDataInfo}
	 * class.
	 * <p>
	 * Implementation Details: In order to copy DataProviders, new Node/EdgeMaps
	 * are created for each of them. Then each DataProvider is asked for the
	 * value of all nodes/edges and the result is stored in the appropriate
	 * Node/Edgemap. Then these Node/Edgemaps are registered as new
	 * DataProviders in the new graph. Therefore, dynamic DataProviders will not
	 * be copied correctly, since the values they return will be made constant
	 * through the copy process.
	 * 
	 * @param graph
	 *            the source graph that is to be duplicated
	 * @param dataInfo
	 *            holds information about which DataProviders should be copied
	 * 
	 * @return the duplicated Graph that has the specified DataProviders
	 *         attatched to it. If <code>graph</code> is <code>null</code>,
	 *         no copying takes place and <code>null</code> is returned.
	 */
	static public Graph DuplicateGraph(Graph graph, GraphDataInfo[] dataInfo) {
		// is there anything to copy ?
		if (graph == null)
			return null;

		Graph copy = new Graph();
		CopyGraphToGraph(graph, copy, dataInfo);

		return copy;
	}

	/**
	 * Creates a copy of the given Graph.
	 * <p>
	 * All nodes and edges of the given graph are copied. The index of a both
	 * nodes and edges is the same in the duplicate graph. The in/out order of
	 * edges is kept as well (! This is not true yet !).
	 * <p>
	 * Furthermore, the user can specify which DataProviders should be copied by
	 * providing their respective keys. For each data provider, the user also
	 * has to specify if it assigns values to nodes or to edges and the type of
	 * the values. This is done via the {@link GraphDataInfo GraphDataInfo}
	 * class.
	 * <p>
	 * Implementation Details: In order to copy DataProviders, new Node/EdgeMaps
	 * are created for each of them. Then each DataProvider is asked for the
	 * value of all nodes/edges and the result is stored in the appropriate
	 * Node/Edgemap. Then these Node/Edgemaps are registered as new
	 * DataProviders in the new graph. Therefore, dynamic DataProviders will not
	 * be copied correctly, since the values they return will be made constant
	 * through the copy process.
	 * 
	 * @param graph
	 *            the source graph that is to be duplicated
	 * @param dataInfo
	 *            holds information about which DataProviders should be copied
	 * 
	 * @return the duplicated Graph that has the specified DataProviders
	 *         attatched to it. If <code>graph</code> is <code>null</code>,
	 *         no copying takes place and <code>null</code> is returned.
	 */
	static public Graph2D ConvertGraphToGraph2D(Graph graph,
			GraphDataInfo[] dataInfo, NodeRealizer defaultNodeRealizer,
			EdgeRealizer defaultEdgeRealizer) {
		// is there anything to copy ?
		if (graph == null)
			return null;

		Graph2D copy = new Graph2D();
		copy.setDefaultNodeRealizer(defaultNodeRealizer);
		copy.setDefaultEdgeRealizer(defaultEdgeRealizer);

		CopyGraphToGraph2D(graph, copy, dataInfo);

		return copy;
	}

	/**
	 * Creates a copy of the given Graph2D.
	 * <p>
	 * All nodes and edges of the given graph are copied. The index of a both
	 * nodes and edges is the same in the duplicate graph. The in/out order of
	 * edges is kept as well (! This is not true yet !).
	 * <p>
	 * Furthermore, the user can specify which DataProviders should be copied by
	 * providing their respective keys. For each data provider, the user also
	 * has to specify if it assigns values to nodes or to edges and the type of
	 * the values. This is done via the {@link GraphDataInfo GraphDataInfo}
	 * class.
	 * <p>
	 * A <code>Graph2D</code> stores Realizers for its Nodes and Edges. Each
	 * Realizer has to provide a <code>createCopy</code> method that creates
	 * an appropriate copy of the Realizer. So, when duplicating a Graph2D, one
	 * could just call these methods. This method takes another approach though.
	 * Instead of calling the <code>createCopy</code> methods directly, it
	 * lets the user specify a {@link RealizerFactory RelizerFactory} that will
	 * handle the copying process for Realizers. For every NodeRealizer in the
	 * source graph,
	 * {@link RealizerFactory#copyNodeRealizer(NodeRealizer) RealizerFactory.copyNodeRealizer}
	 * is called and for every EdgeRealizer in the source graph
	 * {@link RealizerFactory#copyEdgeRealizer(EdgeRealizer) RealizerFactory.copyEdgeRealizer}
	 * is called. This way, the user can provide a specific
	 * <code>RealizerFactory</code> implementation that doesn't simply copy
	 * the Realizers but also modifies them or returns a Realizer type that is
	 * different from the source. (This is extremely useful in the context
	 * adding existing <code>Graph2D</code> objects to a
	 * {@link y25.view.Graph25D Graph25D}. There, all Node- and EdgeRealizers
	 * have to be converted to their 25D counterparts in order to provide
	 * correct OpenGL rendering. This is done internally by passing a
	 * {@link y25.view.realizer.Realizer25DFactory Realizer25DFactory} object to
	 * this routine.)
	 * <p>
	 * Implementation Details: In order to copy DataProviders, new Node/EdgeMaps
	 * are created for each of them. Then each DataProvider is asked for the
	 * value of all nodes/edges and the result is stored in the appropriate
	 * Node/Edgemap. Then these Node/Edgemaps are registered as new
	 * DataProviders in the new graph. Therefore, dynamic DataProviders will not
	 * be copied correctly, since the values they return will be made constant
	 * through the copy process.
	 * 
	 * @param graph
	 *            the source graph that is to be duplicated
	 * @param dataInfo
	 *            holds information about which DataProviders should be copied
	 * @param realizerFactory
	 *            responsible for copying all Node- and EdgeRealizers
	 * 
	 * @return the duplicated Graph2D that has the specified DataProviders
	 *         attatched to it. If <code>graph</code> is <code>null</code>,
	 *         no copying takes place and <code>null</code> is returned.
	 */
	static public Graph2D DuplicateGraph2D(Graph2D graph,
			GraphDataInfo[] dataInfo, RealizerFactory realizerFactory) {
		// is there anything to copy?
		if (graph == null)
			return null;

		Graph2D copy = new Graph2D();
		copy.setDefaultNodeRealizer(realizerFactory.copyNodeRealizer(graph
				.getDefaultNodeRealizer()));
		copy.setDefaultEdgeRealizer(realizerFactory.copyEdgeRealizer(graph
				.getDefaultEdgeRealizer()));

		CopyGraph2DToGraph2D(graph, copy, realizerFactory, dataInfo);

		return copy;
	}

	/**
	 * An interface for creation of node and edge realizers.
	 * 
	 * @author Werner Jainek
	 */
	public interface RealizerFactory {

		/**
		 * Returns a copy of the default node realizer of this factory.
		 */
		public NodeRealizer createDefaultNodeRealizer();

		/**
		 * Returns a copy of the default edge realizer of this factory.
		 */
		public EdgeRealizer createDefaultEdgeRealizer();

		/**
		 * Creates a "25D" version of the passed node realizer as outlined in
		 * the class description above.
		 */
		public NodeRealizer copyNodeRealizer(NodeRealizer nr);

		/**
		 * Creates a "25D" version of the passed edge realizer as outlined in
		 * the class description above.
		 */
		public EdgeRealizer copyEdgeRealizer(EdgeRealizer er);
	}

	/**
	 * Stores information about a <code>DataProvider</code>.
	 * 
	 * @author Werner Jainek
	 */

	static public class GraphDataInfo {

		static public final int NODEMAP = 0;

		static public final int EDGEMAP = 1;

		static public final int BOOLEAN = 10;

		static public final int DOUBLE = 11;

		static public final int INTEGER = 12;

		public Object key;

		public int mapType;

		public int valueType;

		public GraphDataInfo(Object key, int mapType, int valueType) {
			this.key = key;
			this.mapType = mapType;
			this.valueType = valueType;
		}

	}
}
