/**
 * y25 - Visualizing Graphs in 2.5 Dimensions
 * 
 * Arbeitsbereich: Paralleles Rechnen Prof. Kaufmann
 * Studienarbeit:  Werner Jainek
 * Betreuung:      Nina Lehmann
 *
 */
package y25.graphics;

/**
 * Represents an axis-aligned bounding box in three dimensional space.
 * <p>
 * Note that an axis-aligned bounding box is fully specified by knowing the
 * point with the smallest x,y,z-coordinates and the point with the biggest
 * x,y,z-coordinates that still lies inside the bounding box.
 * <p>
 * A newly created bounding box does not bound anything and therefore it
 * <code>isCollapsed()</code>. After one point has been added to the bounding
 * box, it is no longer collapsed, since now it bounds something - the point. It
 * has still no volume though and therefore it <code>isEmpty()</code>. As
 * long as only points that lie on the same plane are added, the bounding box
 * still <code>isEmpty()</code>. Only when the bounding box actually
 * encompasses a region of 3D-space that has a volume, it is no longer empty.
 * 
 * @author Werner Jainek
 */
public class BoundingBox {

	/**
	 * The point with the smallest x,y,z-coordinates that is still inside the
	 * bounding box.
	 */
	protected Point3D min;

	/**
	 * The point with the biggest x,y,z-coordinates that is still inside the
	 * bounding box.
	 */
	protected Point3D max;

	/**
	 * Creates a collapsed bounding box.
	 */
	public BoundingBox() {
		min = new Point3D(Float.POSITIVE_INFINITY, Float.POSITIVE_INFINITY,
				Float.POSITIVE_INFINITY);
		max = new Point3D(Float.NEGATIVE_INFINITY, Float.NEGATIVE_INFINITY,
				Float.NEGATIVE_INFINITY);
	}

	/**
	 * Creates a bounding box that is a copy of the passed bounding box.
	 */
	public BoundingBox(BoundingBox bb) {
		set(bb);
	}

	/**
	 * Creates a bounding box by specifying the points in 3D space with the
	 * smallest and biggest x,y,z-coordinates that are still inside the bounding
	 * box.
	 */
	public BoundingBox(Point3D pmin, Point3D pmax) {
		min = new Point3D(pmin);
		max = new Point3D(pmax);

		float tmp;
		tmp = min.x;
		min.x = Math.min(min.x, max.x);
		max.x = Math.max(tmp, max.x);

		tmp = min.y;
		min.y = Math.min(min.y, max.y);
		max.y = Math.max(tmp, max.y);

		tmp = min.z;
		min.z = Math.min(min.z, max.z);
		max.z = Math.max(tmp, max.z);
	}

	/**
	 * Sets this bounding box to the same extent as the passed bounding box.
	 * <p>
	 * Returns a reference to this bounding box.
	 */
	public BoundingBox set(BoundingBox bb) {
		min = new Point3D(bb.min);
		max = new Point3D(bb.max);
		return this;
	}

	/**
	 * If necessary, enlarges this bounding box to also include the passed
	 * bounding box.
	 * <p>
	 * If the passed bounding box is collapsed, calling this method has no
	 * effect.
	 * <p>
	 * Returns a reference to this bounding box.
	 */
	public BoundingBox include(BoundingBox bb) {
		if (!bb.isCollapsed()) {
			// TODO: this currently needs 12 comparisons -- can be done quicker
			// !
			include(bb.min);
			include(bb.max);
		}
		return this;
	}

	/**
	 * If necessary, enlarges this bounding box to include the passed point.
	 * <p>
	 * Returns a reference to this bounding box.
	 */
	public BoundingBox include(Point3D p) {
		min.x = Math.min(min.x, p.x);
		min.y = Math.min(min.y, p.y);
		min.z = Math.min(min.z, p.z);

		max.x = Math.max(max.x, p.x);
		max.y = Math.max(max.y, p.y);
		max.z = Math.max(max.z, p.z);

		return this;
	}

	/**
	 * If necessary, enlarges this bounding box to include at least one point
	 * that has the passed x-coordinate.
	 * <p>
	 * Returns a reference to this bounding box.
	 */
	public BoundingBox includeX(float x) {
		min.x = Math.min(min.x, x);
		max.x = Math.max(max.x, x);
		return this;
	}

	/**
	 * If necessary, enlarges this bounding box to include at least one point
	 * that has the passed y-coordinate.
	 * <p>
	 * Returns a reference to this bounding box.
	 */
	public BoundingBox includeY(float y) {
		min.y = Math.min(min.y, y);
		max.y = Math.max(max.y, y);
		return this;
	}

	/**
	 * If necessary, enlarges this bounding box to include at least one point
	 * that has the passed z-coordinate.
	 * <p>
	 * Returns a reference to this bounding box.
	 */
	public BoundingBox includeZ(float z) {
		min.z = Math.min(min.z, z);
		max.z = Math.max(max.z, z);
		return this;
	}

	/**
	 * Returns the center of this bounding box.
	 */
	public Point3D getCenter() {
		return Point3D.lerp(min, max, 0.5f);
	}

	/**
	 * Returns the point with the smallest x,y,z-coordinates that still lies
	 * inside the bounding box.
	 */
	public Point3D getMin() {
		return new Point3D(min);
	}

	/**
	 * Returns the point with the biggest x,y,z-coordinates that still lies
	 * inside the bounding box.
	 */
	public Point3D getMax() {
		return new Point3D(max);
	}

	/**
	 * Returns the radius of a bounding sphere that would encompass the entire
	 * bounding box if it was positioned at <code>getCenter()</code>.
	 */
	public float getRadius() {
		return (new Vector3D(min, max)).length() * 0.5f;
	}

	/**
	 * Returns <code>true</code> if this bounding box is collapsed,
	 * <code>false</code> otherwise.
	 */
	public boolean isCollapsed() {
		return (min.x > max.x || min.x > max.x || min.x > max.x);
	}

	/**
	 * Returns this boudning box to a collapsed state.
	 */
	public BoundingBox setCollapsed() {
		min.set(Float.POSITIVE_INFINITY, Float.POSITIVE_INFINITY,
				Float.POSITIVE_INFINITY);
		max.set(Float.NEGATIVE_INFINITY, Float.NEGATIVE_INFINITY,
				Float.NEGATIVE_INFINITY);
		return this;
	}

	/**
	 * Returns <code>true</code> if this bounding box has no volume,
	 * <code>false</code> otherwise.
	 * <p>
	 * Note that a collapsed bounding box also has no volume, so
	 * <code>true</code> is returned in this case as well.
	 */
	public boolean isEmpty() {
		return (min.x == max.x && min.y == max.y && min.z == max.z)
				|| isCollapsed();
	}

	/**
	 * Moves the entire bounding box by the given vector.
	 * <p>
	 * Returns a reference to this bounding box.
	 */
	public BoundingBox translate(Vector3D v) {
		min.translate(v);
		max.translate(v);
		return this;
	}

	/**
	 * Scales the entire bounding box by the given scaling factor.
	 * <p>
	 * Returns a reference to this bounding box.
	 */
	public BoundingBox scale(float s) {
		Point3D center = Point3D.lerp(min, max, 0.5f);
		min.translate(new Vector3D(center, min).scale(s));
		max.translate(new Vector3D(center, max).scale(s));
		return this;
	}

	/**
	 * Adds a border of the specified size around this bounding box.
	 * <p>
	 * Returns a reference to this bounding box.
	 */
	public BoundingBox addBorder(float size) {
		min.x -= size;
		min.y -= size;
		min.z -= size;
		max.x += size;
		max.y += size;
		max.z += size;
		return this;
	}

	public String toString() {
		return "{ min: " + min.toString() + "\n" + "  max: " + max.toString()
				+ " }";
	}
}
