package y25.graphics;

import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;

/**
 * Represents a camera in three dimensional space.
 * 
 * @author Werner Jainek
 */
public class Camera {

	/**
	 * The minimal field of view angle in degrees.
	 */
	public static float minFOV = 0.01f;

	/**
	 * The maximal field of view angle in degrees.
	 */
	public static float maxFOV = 100.0f;

	/**
	 * The field of view angle in degrees.
	 */
	protected float fov; // degrees

	/**
	 * The width of the view area in pixels.
	 */
	protected float width; // pixels

	/**
	 * The height of the view area in pixels.
	 */
	protected float height; // pixels

	/**
	 * The aspect ratio ( width / height ).
	 */
	protected float width_height;

	/**
	 * The z-value of the near clipping plane.
	 */
	protected float zNear;

	/**
	 * The z-value of the far clipping plane.
	 */
	protected float zFar;

	/**
	 * The point this camera looks at.
	 */
	protected Point3D target;

	/**
	 * The camera's rotation around the x-axis. (target beeing the origin of the
	 * coordinate sysyem.)
	 */
	protected float viewRotX;

	/**
	 * The camera's rotation around the y-axis. (target beeing the origin of the
	 * coordinate sysyem.)
	 */
	protected float viewRotY;

	/**
	 * The camera's rotation around the z-axis. (target beeing the origin of the
	 * coordinate sysyem.)
	 */
	protected float viewRotZ;

	/**
	 * The camera's distance from the target.
	 */
	protected float viewDistance;

	/**
	 * The transform that transforms world coordinates into camera coordinates.
	 */
	protected Transform worldToCamera;

	/**
	 * The transform that transforms camera coordinates into (NDC) normalized
	 * device coordinates.
	 */
	protected Transform cameraToNDC;

	/**
	 * The transform that transforms NDC coordinates into screen soordinates.
	 */
	protected Transform ndcToScreen;

	/**
	 * A shortcut transform that transforms world coordinates directly into
	 * screen soordinates.
	 */
	protected Transform worldToScreen;

	/**
	 * An indicator that settings of the camera changed. Set to
	 * <code>false</code> whenever <code>updateWorldToCamera</code> is
	 * called.
	 */
	private boolean cameraChanged;

	/**
	 * Creates a default camera. The following settings are used:
	 * <ul>
	 * <li>field of view - 60 degrees</li>
	 * <li>width/height - 200 pixels</li>
	 * <li>near clipping plane - 0.1</li>
	 * <li>far clipping plane - 3000.0</li>
	 * <li>target - (0,0,0)</li>
	 * <li>rotationX/Y/Z - 0</li>
	 * <li>view distance - 500.0f</li>
	 * </ul>
	 */
	public Camera() {
		fov = 60.0f;
		width = 200.0f;
		height = 200.0f;
		width_height = 1.0f;
		zNear = 0.1f;
		zFar = 3000.0f;

		target = new Point3D(0, 0, 0);
		setRotation(0, 0, 0);
		viewDistance = 500.0f;

		worldToCamera = new Transform();
		cameraToNDC = new Transform();
		ndcToScreen = new Transform();
		worldToScreen = new Transform();
		updateWorldToCameraToNDCToScreen();

	}

	/**
	 * Recalculates the world to camera transform and the camera to NDC
	 * (normalized device coordinates) transform from the camera's parameters.
	 * Then sets <code>cameraChanged</code> to false, since the current
	 * transform reflects the camera's settings from now on.
	 */
	private void updateWorldToCameraToNDCToScreen() {

		// WORLD TO CAMERA
		worldToCamera.setIdentity();
		Transform.translate(new Vector3D(target).reverse())
				.apply(worldToCamera);
		Transform.rotateZ(-viewRotZ).apply(worldToCamera);
		Transform.rotateX(-viewRotX).apply(worldToCamera);
		Transform.rotateY(-viewRotY).apply(worldToCamera);
		Transform.translate(new Vector3D(0, -viewDistance, 0)).apply(
				worldToCamera);
		Transform.rotateX(90).apply(worldToCamera);

		// CAMERA TO NDC
		cameraToNDC = Transform.perspective(fov, width_height, zNear, zFar);

		// NDC TO SCREEN
		ndcToScreen = Transform.viewport(0, 0, width, height);

		// WORLD TO SCREEN
		worldToScreen.setIdentity();
		worldToCamera.apply(worldToScreen);
		cameraToNDC.apply(worldToScreen);
		ndcToScreen.apply(worldToScreen);

		cameraChanged = false;
	}

	/**
	 * Sets the passed drawable's OpenGL projection and modelview matrices to
	 * this camera's settings.
	 * <p>
	 * After calling this method, world geometry will be drawn as if viewed from
	 * this camera.
	 */
	public void activateProjection(GLAutoDrawable drawable) {
		GL gl = drawable.getGL();

		// if camera did change, recalculate the transformations
		if (cameraChanged) {
			updateWorldToCameraToNDCToScreen();
		}

		float[] vpm = ndcToScreen.getArray(Transform.COLUMNS);
		gl.glViewport((int) (vpm[12] / vpm[0] - 1),
				(int) (vpm[13] / vpm[5] - 1), 2 * (int) vpm[0],
				2 * (int) vpm[5]);
		// same as gl.glViewport(0,0,(int)width,(int)height)!, but these values
		// are extracted so that the computation is only done ince in
		// updateWorldToCam....()

		gl.glMatrixMode(GL.GL_PROJECTION);
		gl.glLoadMatrixf(cameraToNDC.getArray(Transform.COLUMNS), 0);

		gl.glMatrixMode(GL.GL_MODELVIEW);
		gl.glLoadMatrixf(worldToCamera.getArray(Transform.COLUMNS), 0);
	}

	/**
	 * Returns the screen coordinates of the passed point.
	 */
	public Point3D getScreenCoordinates(Point3D p) {
		Point3D result = new Point3D(p);
		result.z = -result.z;

		worldToScreen.apply(result);
		return result;
	}

	/**
	 * Activates a "Head Up Display" for the given drawable.
	 * <p>
	 * After calling this method, coordinates retrieved via
	 * <code>glRasterPos</code> or <code>getHUDCoordinates</code> can be
	 * passed to <code>glVertex</code> and they will be rendered to the
	 * correct position.
	 * <p>
	 * Every call of this method has to be matched by a call to
	 * <code>deactivateHUD</code>.
	 */
	public void activateHUD(GLAutoDrawable drawable) {
		GL gl = drawable.getGL();

		gl.glMatrixMode(GL.GL_PROJECTION);
		gl.glPushMatrix();
		gl.glLoadIdentity();
		gl.glOrtho(0.0, // left edge is zero
				width, // right edge is width
				0.0, // bottom edge is zero
				height, // top edge is height
				0.0, -1.0); // near and far clipping planes.

		gl.glMatrixMode(GL.GL_MODELVIEW);
		gl.glPushMatrix();
		gl.glLoadIdentity();
	}

	/**
	 * Deactivates the "Head Up Display" and returns to the previous projection
	 * settings.
	 * <p>
	 * It is an error to call this method before <code>activateHUD</code> has
	 * been called. It is also an error not to match a call of
	 * <code>activateHUD</code> with a call to this method.
	 */
	public void deactivateHUD(GLAutoDrawable drawable) {
		GL gl = drawable.getGL();

		gl.glMatrixMode(GL.GL_PROJECTION);
		gl.glPopMatrix();

		gl.glMatrixMode(GL.GL_MODELVIEW);
		gl.glPopMatrix();
	}

	/**
	 * Returns the HUD-Coordinates of the passed point.
	 * <p>
	 * If the passed point lies outside of the viewing area, <code>null</code>
	 * is returned.
	 * <p>
	 * Before using the retrieved coordinates to draw elements on the screen,
	 * <code>activateHUD</code> has to be called.
	 */
	public Point3D getHUDCoordinates(GLAutoDrawable drawable, Point3D p) {
		GL gl = drawable.getGL();
		gl.glRasterPos3f(p.x, p.y, -p.z); // TODO: check if (-) here is
		// correct
		float rasterPos[] = new float[4];
		gl.glGetFloatv(GL.GL_CURRENT_RASTER_POSITION, rasterPos, 0);
		byte rasterPosValid[] = new byte[1];
		gl
				.glGetBooleanv(GL.GL_CURRENT_RASTER_POSITION_VALID,
						rasterPosValid, 0);
		if (rasterPosValid[0] == 0) {
			return null;
		} else {
			return new Point3D(rasterPos[0], rasterPos[1], rasterPos[2]);
		}
	}

	// returns a vector that points to the camera (all coords relative to
	// current opengl coord. system)

	/**
	 * Returns a vector that points to the camera.
	 * <p>
	 * Note that this method is different from calculating a vector from
	 * <code>from</code> to <code>getPosition()</code>, since the passed
	 * point <code>from</code> is interpreted with respect to the current
	 * OpenGL matrix stack. This makes it possible to make multiple calls to
	 * <code>glTranslate</code>, <code>glScale</code>, etc. and to then
	 * retrieve a vector to the camera without having to transform the local
	 * coordinates to world coordinates first.
	 */
	public Vector3D toCamera(GLAutoDrawable drawable, Point3D from) {
		GL gl = drawable.getGL();

		// transform model point to world coordinates
		Point3D modelPoint = new Point3D(from);
		modelPoint.z = -modelPoint.z;

		// apply modelview transformation (go from model to world to camera
		// coordinates)
		float mv[] = new float[16];
		gl.glGetFloatv(GL.GL_MODELVIEW_MATRIX, mv, 0);
		Transform t = new Transform(mv, Transform.COLUMNS);
		t.apply(modelPoint);

		// undo view transformation (go from camera to world coordinates)
		worldToCamera.getInverted().apply(modelPoint);

		// flip z-axis
		modelPoint.z = -modelPoint.z;

		// compute toCam vector (from modelOrigin to position)
		Vector3D toCam = new Vector3D(modelPoint, getPosition());

		return toCam;
	}

	/**
	 * Returns the view direction of the camera.
	 */
	public Vector3D getViewDirection() {
		Point3D p = getPosition();
		Point3D t = getTarget();
		return new Vector3D(p, t).normalize();
	}

	/**
	 * Returns the vector in three dimensional space that corresponds to the
	 * upward direction in the viewing plane.
	 */
	public Vector3D getUpDirection() {
		if (cameraChanged)
			updateWorldToCameraToNDCToScreen();

		Vector3D up = new Vector3D(0, 0, 1);
		worldToCamera.apply(up);
		up.z = -up.z;
		return up.normalize();
	}

	/**
	 * Sets the field of view of this camera in angles.
	 */
	public void setFOV(float f) {
		f = Math.max(f, 179.0f);
		f = Math.min(f, 1.0f);
		fov = f;
		cameraChanged = true;
	}

	/**
	 * Returns the field of view of this camera in angles.
	 */
	public float getFOV() {
		return fov;
	}

	/**
	 * Recalculates the aspect ratio from the camera's viewing plane width and
	 * height.
	 */
	private void updateAspectRatio() {
		width_height = Math.abs(width / height);
	}

	/**
	 * Sets the width of the camera's viewing plane in pixels.
	 */
	public void setWidth(float w) {
		width = w;
		updateAspectRatio();
		cameraChanged = true;
	}

	/**
	 * Sets the hegiht of the camera's viewing plane in pixels.
	 */
	public void setHeight(float h) {
		height = h;
		updateAspectRatio();
		cameraChanged = true;
	}

	/**
	 * Returns the camaera's aspect ratio.
	 */
	public float getAspectRatio() {
		return width_height;
	}

	/**
	 * Adjusts the clipping planes of this camera so that the content of the
	 * passed bounding box can be displayed correctly.
	 * <p>
	 * If the passed bounding box is collapsed, this method has no effect.
	 */
	public void setClippingPlanes(BoundingBox bb) {

		if (!bb.isCollapsed()) {
			Point3D camPos = getPosition();
			camPos.z = -camPos.z;

			Point3D bbCenter = bb.getCenter();
			bbCenter.z = -bbCenter.z;

			// get projected distance of bbCenter from camerea
			Vector3D viewDir = new Vector3D(camPos, target).normalize();
			Vector3D toBBCenter = new Vector3D(camPos, bbCenter);
			float projectedDist = viewDir.dot(toBBCenter);

			// conservative approximation of bounding box
			float bbRadius = bb.getRadius();

			zNear = (float) Math.max(0.01, projectedDist - bbRadius);
			zFar = (float) Math.max(0.011, projectedDist + bbRadius);

			cameraChanged = true;
		}
	}

	/**
	 * Sets the near and far clipping planes to the specified values.
	 */
	public void setClippingPlanes(float zN, float zF) {
		zNear = Math.max(zN, 0.0001f);
		zFar = Math.max(zF, zNear + 0.0001f);
		cameraChanged = true;
	}

	/**
	 * Sets the near clipping plane of this camera.
	 */
	public void setNearClippingPlane(float near) {
		zNear = near;
		cameraChanged = true;
	}

	/**
	 * Sets the far clipping plane of this camera.
	 */
	public void setFarClippingPlane(float far) {
		zFar = far;
		cameraChanged = true;
	}

	/**
	 * Returns the z-value of the near clipping plane.
	 */
	public float getNearClippingPlane() {
		return zNear;
	}

	/**
	 * Returns the z-value of the far clipping plane.
	 */
	public float getFarClippingPlane() {
		return zFar;
	}

	/**
	 * Sets the position of this camera.
	 * <p>
	 * The camera's target remains at the same place.
	 */
	public void setPosition(Point3D pos) {
		Point3D p = new Point3D(pos);
		p.z = -p.z;

		Vector3D dir = new Vector3D(target, p);

		if (dir.isZero())
			return; // if position is to be set on the target, we ignore that !

		viewDistance = dir.length();
		dir.normalize();

		// phi/theta=0 starts at the y-axis
		// theta [0..pi] leads to +z-values
		// theta [-pi..0] leads to -z values
		// phi=90 is at x axis, phi=180 is at -y, phi=270 is at -x
		float theta = (float) Math.asin(-dir.z);
		float smallradius = (float) Math.abs(Math.cos(theta));
		theta = (float) Math.toDegrees(theta);

		float ph_s = (float) Math.asin(Math.min(dir.x / smallradius, 1.0));
		float ph_c = (float) Math.acos(Math.min(dir.y / smallradius, 1.0));

		float phi = 0.0f;
		if (ph_s >= 0) {
			phi = (float) Math.toDegrees(ph_c);
		} else {
			phi = 360.0f - (float) Math.toDegrees(ph_c);
		}

		viewRotX = -theta;
		viewRotZ = -phi;

		cameraChanged = true;
	}

	/**
	 * Returns the position of the camera.
	 */
	public Point3D getPosition() {
		Point3D pos = new Point3D(0, viewDistance, 0);
		// Transform.rotateY(viewRotY).apply(pos); // we don't need to this as
		// this is only affecting the tilt of the camera, not its position
		Transform.rotateX(viewRotX).apply(pos);
		Transform.rotateZ(viewRotZ).apply(pos);
		pos.x += target.x;
		pos.y += target.y;
		pos.z += target.z;
		pos.z = -pos.z; // outside world works with flipped z-coordinates !
		return pos;
	}

	/**
	 * Sets the viewing target of this camera to the passed x,y,z coordinates.
	 */
	public void setTarget(float x, float y, float z) {
		target.x = x;
		target.y = y;
		target.z = -z;
		cameraChanged = true;
	}

	/**
	 * Sets the viewing target of this camera to the passed point.
	 */
	public void setTarget(Point3D p) {
		target.x = p.x;
		target.y = p.y;
		target.z = -p.z;
		cameraChanged = true;
	}

	/**
	 * Returns the viweing target of this camera.
	 */
	public Point3D getTarget() {
		Point3D t = new Point3D(target);
		t.z = -t.z;
		return t;
	}

	/**
	 * Sets the x,y,z rotation of this camera. (In this context,
	 * <code>target</code> is the center of the coordinate system that is used
	 * for rotation.)
	 */
	public void setRotation(float x, float y, float z) {
		// TODO: -/+ the rotations to match with usual logic
		viewRotX = x;
		viewRotY = y;
		viewRotZ = z;
		cameraChanged = true;
	}

	/**
	 * Returns the camera's rotation around the x-axis.
	 */
	public float getRotationX() {
		return viewRotX;
	}

	/**
	 * Returns the camera's rotation around the y-axis.
	 */
	public float getRotationY() {
		return viewRotY;
	}

	/**
	 * Returns the camera's rotation around the z-axis.
	 */
	public float getRotationZ() {
		return viewRotZ;
	}

	/**
	 * Sets the view distance of this camera to the target.
	 */
	public void setViewDistance(float viewDist) {
		viewDistance = Math.abs(viewDist);
		cameraChanged = true;
	}

	/**
	 * Returns the distance from this camera to the target.
	 */
	public float getViewDistance() {
		return viewDistance;
	}

	/**
	 * Rotates the camera by the passed angles in degrees.
	 */
	public void rotate(float x, float y, float z) {
		viewRotX += x;
		viewRotY += y;
		viewRotZ += z;
		cameraChanged = true;
	}

	/**
	 * Moves the camera with respect to it's viewing direction. A factor of
	 * <code>0.1</code> means getting 10 percent closer to the target, a
	 * factor of <code>-0.1</code> means moving 10 percent farther away from
	 * the target.
	 * <p>
	 * A factor of greater than <code>1</code> is not allowed.
	 */
	public void zoom(float factor) {
		factor = 1.0f - factor;
		if (factor < 0.01f)
			factor = 0.01f; // make sure we scale by a reasonable amount (and
		// not negative!)
		viewDistance *= factor;
		cameraChanged = true;
	}

	// 180deg == overly perspective
	// 0 deg == orthogonal
	/**
	 * Sets the field of view of this camera while keeping the viewing area as
	 * coherent as possible with the previous setting.
	 * <p>
	 * A value of <code>180</code> degrees for the parameter
	 * <code>angle</code> means an overly perspective projection, whereas a
	 * value of <code>0</code> degrees means an orthogonal view of the scene.
	 */
	public void setPerspective(float angle) {
		float fovDifference = angle - fov;
		changePerspective(fovDifference);
	}

	/**
	 * Increases or decreases the field of viwe by the given angle while keeping
	 * the viewing area as coherent as possible with the previous setting.
	 * <p>
	 * When <code>angle</code> is positive, the 'perspectiveness' gets
	 * stronger. To compensate for the bigger field of view the camera moves
	 * closer to the target.
	 * <p>
	 * When <code>angle</code> is negative, the 'perspectiveness' gets waeker.
	 * To comensate for the smaller field of view the camera moves further away
	 * from the target.
	 * <p>
	 * Note that repeatedly calling this method with small angle-steps leads to
	 * the famous 'Hitchcock'-effect.
	 * 
	 * @param angle
	 */
	public void changePerspective(float angle) {

		// compute new fov and cap it
		float fovNew = fov + angle;
		if (fovNew >= maxFOV)
			fovNew = maxFOV;
		if (fovNew <= minFOV)
			fovNew = minFOV;

		// compute the half fov
		float halfOld = fov / 2.0f;
		float halfNew = fovNew / 2.0f;

		// compute the height of the current fov, this is to remain fixed
		float fixedHeight = (float) Math.atan(Math.toRadians(halfOld))
				* viewDistance;

		// compute the new viewDistance based on the new fov and the fixed
		// height
		float viewDistanceNew = fixedHeight
				/ (float) Math.atan(Math.toRadians(halfNew));

		// set them
		fov = fovNew;
		viewDistance = viewDistanceNew;
		cameraChanged = true;
	}

	public String toString() {
		return "[" + "\n" + "   fov " + String.valueOf(fov) + "\n"
				+ "   aspect(width/height) " + String.valueOf(width_height)
				+ "\n" + "   zNear " + String.valueOf(zNear) + "\n"
				+ "   zFar " + String.valueOf(zFar) + "\n" + "   target "
				+ target.toString() + "\n" + "   rotation(XYZ) "
				+ String.valueOf(viewRotX) + " " + String.valueOf(viewRotY)
				+ " " + String.valueOf(viewRotZ) + "\n" + "   distance "
				+ String.valueOf(viewDistance) + "\n" + "]";
	}
}
