/**
 * y25 - Visualizing Graphs in 2.5 Dimensions
 * 
 * Arbeitsbereich: Paralleles Rechnen Prof. Kaufmann
 * Studienarbeit:  Werner Jainek
 * Betreuung:      Nina Lehmann
 *
 */
package y25.graphics;

import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;

/**
 * A utility class for drawing three dimensional primitives.
 * 
 * @author Werner Jainek
 */
public class Graphics3D {

	/**
	 * Draws a cylinder.
	 * 
	 * @param drawable
	 *            the drawable that this drawing operations go to.
	 * @param fromm
	 *            the start position of the cylinder.
	 * @param too
	 *            the end position of the cylinder.
	 * @param radius
	 *            the radius of the cylinder.
	 * @param sides
	 *            the number of sides the cylinder will have. It must have at
	 *            least 3 sides.
	 * @param capped
	 *            determines if a cap should be drawn at the top and at the
	 *            bottom of the cyliner.
	 */
	public static void drawCylinder(GLAutoDrawable drawable, Point3D fromm,
			Point3D too, float radius, int sides, boolean capped) {

		GL gl = drawable.getGL();

		// flip z-values
		Point3D from = new Point3D(fromm);
		from.z = -from.z;
		Point3D to = new Point3D(too);
		to.z = -to.z;

		// check arguments
		if (radius < 0)
			radius = 0.01f;
		if (sides < 3)
			sides = 3;

		// create the local coordinate system
		Vector3D fromto = new Vector3D(from, to);
		Vector3D dir = fromto.getNormalized();
		Vector3D up = (dir.x != 0 || dir.y != 0) ? (new Vector3D(0, 0, -1))
				: (new Vector3D(1, 0, 0));
		Vector3D right = dir.getCrossed(up).normalize();
		up = right.getCrossed(dir);

		// create the quads & caps that compose the cylinder

		// calculate starting point (we start at the bottom since this looks
		// best if we only have 3 sides)
		float angle = (float) (0.75 * 2.0 * Math.PI);
		Vector3D previousNormal = new Vector3D(up.getReversed());
		Point3D previous = new Point3D(from);
		previous.translate(up.getScaled(-radius));

		// run through all sides of the cylinder
		float step = (float) (2.0 * Math.PI) / (float) sides;
		for (int i = 0; i < sides; ++i) {

			// determine next point on the circle
			angle -= step;
			Vector3D nextNormal = new Vector3D(0, 0, 0);
			nextNormal.add(right.getScaled((float) Math.cos(angle)));
			nextNormal.add(up.getScaled((float) Math.sin(angle)));

			Point3D next = new Point3D(from);
			next.translate(nextNormal.getScaled(radius));

			// create the quad for the current side
			gl.glBegin(GL.GL_QUADS);
			gl.glNormal3f(previousNormal.x, previousNormal.y, previousNormal.z);
			gl.glVertex3f(previous.x, previous.y, previous.z);

			gl.glNormal3f(nextNormal.x, nextNormal.y, nextNormal.z);
			gl.glVertex3f(next.x, next.y, next.z);

			gl.glNormal3f(nextNormal.x, nextNormal.y, nextNormal.z);
			gl.glVertex3f(next.x + fromto.x, next.y + fromto.y, next.z
					+ fromto.z);

			gl.glNormal3f(previousNormal.x, previousNormal.y, previousNormal.z);
			gl.glVertex3f(previous.x + fromto.x, previous.y + fromto.y,
					previous.z + fromto.z);
			gl.glEnd();

			// create the triangles for the caps
			if (capped) {
				gl.glBegin(GL.GL_TRIANGLES);
				gl.glNormal3f(-fromto.x, -fromto.y, -fromto.z);
				gl.glVertex3f(from.x, from.y, from.z);
				gl.glVertex3f(next.x, next.y, next.z);
				gl.glVertex3f(previous.x, previous.y, previous.z);

				gl.glNormal3f(fromto.x, fromto.y, fromto.z);
				gl.glVertex3f(to.x, to.y, to.z);
				gl.glVertex3f(previous.x + fromto.x, previous.y + fromto.y,
						previous.z + fromto.z);
				gl.glVertex3f(next.x + fromto.x, next.y + fromto.y, next.z
						+ fromto.z);
				gl.glEnd();
			}

			// advance
			previous = next;
			previousNormal = nextNormal;
		}

	}

	/**
	 * Draws a cylinder with a tilted top and bottom.
	 * <p>
	 * The tilt is determined by normals at the top and bottom of the cylinder.
	 * 
	 * @param drawable
	 *            the drawable that this drawing operations go to.
	 * @param fromDir
	 *            the normal that determines the tilt at the bottom of the
	 *            cylinder.
	 * @param from
	 *            the start (bottom) position of the cylinder.
	 * @param to
	 *            the end (top) position of the cylinder.
	 * @param toDir
	 *            the normal that determines the tilt at the top of the
	 *            cylinder.
	 * @param radius
	 *            the radius of the cylinder.
	 * @param sides
	 *            the number of sides the cylinder will have. It must have at
	 *            least 3 sides.
	 * @param capped
	 *            determines if a cap should be drawn at the top and at the
	 *            bottom of the cyliner.
	 * 
	 */
	public static void drawCylinder(GLAutoDrawable drawable, Vector3D fromDir,
			Point3D from, Point3D to, Vector3D toDir, float radius, int sides,
			boolean capped) {
		GL gl = drawable.getGL();

		// check arguments
		if (radius < 0)
			radius = 0.01f;
		if (sides < 3)
			sides = 3;
		Vector3D fromdir = fromDir.getNormalized();
		Vector3D todir = toDir.getNormalized();

		// create the local coordinate system
		Vector3D fromto = new Vector3D(from, to);
		Vector3D dir = fromto.getNormalized();
		Vector3D up = (dir.x != 0 || dir.y != 0) ? (new Vector3D(0, 0, -1))
				: (new Vector3D(1, 0, 0));
		Vector3D right = dir.getCrossed(up).normalize();
		up = right.getCrossed(dir);

		// make sure normals of cutting planes point outward !
		// also, if cutting plane-normals are orthogonal to fromto, don't use
		// them
		Vector3D fromtonormalized = fromto.getNormalized();
		float projectionFromDir = fromtonormalized.dot(fromdir);
		float projectionToDir = fromtonormalized.dot(todir);

		if (Math.abs(projectionFromDir) < 0.01f) { // plane-normal is
			// orthogonal
			fromdir.set(fromtonormalized).reverse();
		} else if (projectionFromDir > 0.0f) { // plane-normal is wrong
			// direction
			fromdir.reverse();
		}

		if (Math.abs(projectionToDir) < 0.01f) { // plane-normal is
			// orthogonal
			todir.set(fromtonormalized);
		} else if (projectionToDir < 0.0f) { // plane-normal is wrong
			// direction
			todir.reverse();
		}

		// create the quads & caps that compose the cylinder

		// calculate starting point (we start at the bottom since this looks
		// best if we only have 3 sides)
		float angle = (float) (0.75 * 2.0 * Math.PI);
		Vector3D previousNormal = new Vector3D(up.getReversed());
		Point3D previous = new Point3D(from);
		previous.translate(up.getScaled(-radius));

		// run through all sides of the cylinder
		float step = (float) (2.0 * Math.PI) / (float) sides;
		for (int i = 0; i < sides; ++i) {

			// determine next point on the circle
			angle -= step;
			Vector3D nextNormal = new Vector3D(0, 0, 0);
			nextNormal.add(right.getScaled((float) Math.cos(angle)));
			nextNormal.add(up.getScaled((float) Math.sin(angle)));

			Point3D next = new Point3D(from);
			next.translate(nextNormal.getScaled(radius));

			// project the calculated points ono the respective planes (F,T)
			Point3D prF = project(previous, fromto, from, fromdir);
			Point3D prT = project(previous, fromto, to, todir);
			Point3D neF = project(next, fromto, from, fromdir);
			Point3D neT = project(next, fromto, to, todir);

			// create the quad for the current side
			gl.glBegin(GL.GL_QUADS);
			gl.glNormal3f(previousNormal.x, previousNormal.y, previousNormal.z);
			gl.glVertex3f(prF.x, prF.y, prF.z);

			gl.glNormal3f(nextNormal.x, nextNormal.y, nextNormal.z);
			gl.glVertex3f(neF.x, neF.y, neF.z);

			gl.glNormal3f(nextNormal.x, nextNormal.y, nextNormal.z);
			gl.glVertex3f(neT.x, neT.y, neT.z);

			gl.glNormal3f(previousNormal.x, previousNormal.y, previousNormal.z);
			gl.glVertex3f(prT.x, prT.y, prT.z);
			gl.glEnd();

			// create the triangles for the caps
			if (capped) {
				gl.glBegin(GL.GL_TRIANGLES);
				gl.glNormal3f(fromdir.x, fromdir.y, fromdir.z);
				gl.glVertex3f(from.x, from.y, from.z);
				gl.glVertex3f(neF.x, neF.y, neF.z);
				gl.glVertex3f(prF.x, prF.y, prF.z);

				gl.glNormal3f(todir.x, todir.y, todir.z);
				gl.glVertex3f(to.x, to.y, to.z);
				gl.glVertex3f(prT.x, prT.y, prT.z);
				gl.glVertex3f(neT.x, neT.y, neT.z);
				gl.glEnd();
			}

			// advance
			previous = next;
			previousNormal = nextNormal;
		}

	}

	/**
	 * Draws a cone.
	 * 
	 * @param drawable
	 *            the drawable that this drawing operations go to.
	 * @param tipp
	 *            the tip of the cone.
	 * @param basee
	 *            the base position of the cone.
	 * @param radius
	 *            the radius of the base.
	 * @param sides
	 *            the number of sides the cone will have. It must have at least
	 *            3 sides.
	 * @param capped
	 *            determines if a cap should be drawn at the base of the cone.
	 */
	public static void drawCone(GLAutoDrawable drawable, Point3D tipp,
			Point3D basee, float radius, int sides, boolean capped) {

		GL gl = drawable.getGL();

		// flip z-values
		Point3D tip = new Point3D(tipp);
		tip.z = -tip.z;
		Point3D base = new Point3D(basee);
		base.z = -base.z;

		Vector3D tipToBase = new Vector3D(tip, base);

		// compute local coordinate system
		Point3D center = tip.getTranslated(tipToBase);
		Vector3D right = tipToBase.getOrthogonal().normalize();
		Vector3D up = tipToBase.getCrossed(right).normalize();

		// compute angle-step
		float angleStep = (float) (Math.PI * 2.0 / sides);

		// TODO: Normal computation is not correct yet, but the error should be
		// close to unnoticeable

		float angle = angleStep;
		Point3D current = center.getTranslated(right.getScaled(radius));
		Vector3D currentNormal = new Vector3D(right);
		for (int i = 0; i < sides; ++i) {

			// compute coordinates of the current vertex on the circle
			Vector3D nextNormal = new Vector3D(0, 0, 0);
			nextNormal.add(right.getScaled((float) Math.cos(angle)));
			nextNormal.add(up.getScaled((float) Math.sin(angle)));
			Point3D next = new Point3D(center);
			next.translate(nextNormal.getScaled(radius));

			// triangle to the tip
			gl.glBegin(GL.GL_TRIANGLES);

			gl.glNormal3f(currentNormal.x, currentNormal.y, currentNormal.z);
			gl.glVertex3f(current.x, current.y, current.z);
			gl.glNormal3f(currentNormal.x, currentNormal.y, currentNormal.z);
			gl.glVertex3f(tip.x, tip.y, tip.z);
			gl.glNormal3f(nextNormal.x, nextNormal.y, nextNormal.z);
			gl.glVertex3f(next.x, next.y, next.z);

			// triangle for capping
			if (capped) {
				gl.glNormal3f(tipToBase.x, tipToBase.y, tipToBase.z);
				gl.glVertex3f(current.x, current.y, current.z);
				gl.glVertex3f(next.x, next.y, next.z);
				gl.glVertex3f(center.x, center.y, center.z);
			}

			gl.glEnd();

			// advance
			current = next;
			currentNormal = nextNormal;
			angle += angleStep;
		}
	}

	/**
	 * Projects a point <code>p</code> along the direction <code>dir</code>
	 * onto the plane that is defined by (<code>origin</code>,<code>normal</code>).
	 * The normal has to be of unit length.
	 * <p>
	 * Returns the projected point.
	 */
	public static Point3D project(Point3D p, Vector3D dir, Point3D origin,
			Vector3D normal) {

		// determine the orthogonal distance of p to the plane ( n * (p-origin)
		// )
		float distance = normal.dot(new Vector3D(origin, p));

		// determine the orthogonal distance we would travel if we added dir to
		// p
		float traveldistance = -normal.dot(dir);

		// how many times do we have to travel along dir to reach the plane ?
		float travelsteps = distance / traveldistance;

		// compute the projected point
		Point3D projected = new Point3D(p);
		projected.translate(dir.getScaled(travelsteps));

		return projected;
	}

}
