/**
 * y25 - Visualizing Graphs in 2.5 Dimensions
 * 
 * Arbeitsbereich: Paralleles Rechnen Prof. Kaufmann
 * Studienarbeit:  Werner Jainek
 * Betreuung:      Nina Lehmann
 *
 */
package y25.graphics;

/**
 * Represents a point in three dimensional space.
 * <p>
 * A point's x,y and z coordinates can be retrieved and set by accessing it's
 * x,y and z members directly, e.g. <code>point.x = 3*point.z</code>.
 * <p>
 * Most operations on a <code>Point3D</code> can be chained, since a reference
 * to the point is returned by most methods, e.g.
 * <code>point.set(2,3,4).translate(-1,2,3)</code>.
 * 
 * @author Werner Jainek
 */
public class Point3D {

	/**
	 * The x-coordinate of the point.
	 */
	public float x;

	/**
	 * The y-coordinate of the point.
	 */
	public float y;

	/**
	 * The z-coordinate of the point.
	 */
	public float z;

	/**
	 * Creates a point at <code>(0,0,0)</code>.
	 */
	public Point3D() {
		x = y = z = 0;
	}

	/**
	 * Creates a point at the given coordinates.
	 */
	public Point3D(float xx, float yy, float zz) {
		set(xx, yy, zz);
	}

	/**
	 * Creates a point at the same position as the passed point.
	 */
	public Point3D(Point3D p) {
		set(p);
	}

	/**
	 * Creates a point by interpreting the passed vector as a position vector.
	 * 
	 */
	public Point3D(Vector3D v) {
		set(v);
	}

	/**
	 * Sets this point to the given coordinates and returns a reference to this
	 * point.
	 */
	public Point3D set(float xx, float yy, float zz) {
		x = xx;
		y = yy;
		z = zz;
		return this;
	}

	/**
	 * Sets this point to the same coordinates as the passed one and returns a
	 * reference to this point.
	 */
	public Point3D set(Point3D p) {
		x = p.x;
		y = p.y;
		z = p.z;
		return this;
	}

	/**
	 * Sets this point to the same coordinates as the passed vector and returns
	 * a reference to this point.
	 */
	public Point3D set(Vector3D v) {
		x = v.x;
		y = v.y;
		z = v.z;
		return this;
	}

	/**
	 * Returns <code>true</code> if this point lies at the origin,
	 * <code>false</code> otherwise.
	 */
	public boolean isZero() {
		return x == 0 && y == 0 && z == 0;
	}

	/**
	 * Translates this point by the given vector. In mathematical notation, this
	 * would be <code>point = point + vector</code>.
	 */
	public Point3D translate(Vector3D v) {
		x += v.x;
		y += v.y;
		z += v.z;
		return this;
	}

	/**
	 * Creates a copy of this point, translates it by the given vector and
	 * returns the copy.
	 */
	public Point3D getTranslated(Vector3D v) {
		Point3D p = new Point3D(this);
		return p.translate(v);
	}

	public String toString() {
		return "[" + String.valueOf(x) + " " + String.valueOf(y) + " "
				+ String.valueOf(z) + "]";
	}

	/**
	 * Performs a linear interpolation between two points.
	 * <p>
	 * If <code>amount</code> is <code>0</code>, a new point at position
	 * <code>a</code> is returned, if <code>amount</code> is <code>1</code>,
	 * a point at position <code>b</code> is returned. For values between
	 * <code>0</code> and <code>1</code>, <code>a</code> and
	 * <code>b</code> are interpolated. For values in the range
	 * <code>(-inf,0)</code> and <code>(1,inf)</code>, <code>a</code> and
	 * <code>b</code> are extrapolated.
	 */
	public static Point3D lerp(Point3D a, Point3D b, float amount) {
		Vector3D dir = new Vector3D(a, b);
		return new Point3D(a).translate(dir.scale(amount));
	}

	/**
	 * Returns a <code>Vector3D</code> with the same x,y,z entries as this
	 * point.
	 */
	public Vector3D getVector3D() {
		return new Vector3D(x, y, z);
	}
}