/**
 * y25 - Visualizing Graphs in 2.5 Dimensions
 * 
 * Arbeitsbereich: Paralleles Rechnen Prof. Kaufmann
 * Studienarbeit:  Werner Jainek
 * Betreuung:      Nina Lehmann
 *
 */
package y25.graphics;

/**
 * Represents a vector in three dimensional space.
 * 
 * A vector's x,y and z coordinates can be retrieved and set by accessing it's
 * x,y and z members directly, e.g. <code>vector.x = 3*vector.z</code>.
 * <p>
 * Most operations on a <code>Vector3D</code> can be chained, since a
 * reference to the vector is returned by most methods, e.g.
 * <code>vector.set(2,3,4).add(-1,2,3).scale(0.5).reverse()</code>.
 * 
 * @author Werner Jainek
 */
public class Vector3D {

	/**
	 * The x-coordinate of the vector.
	 */
	public float x;

	/**
	 * The y-coordinate of the vector.
	 */
	public float y;

	/**
	 * The z-coordinate of the vector.
	 */
	public float z;

	/**
	 * Creates a null-vector (a vector whose entries are all 0).
	 */
	public Vector3D() {
		x = y = z = 0;
	}

	/**
	 * Creates a vector that points in the same direction as the passed vector.
	 */
	public Vector3D(Vector3D v) {
		set(v);
	}

	/**
	 * Creates a vector that points in the given direction.
	 */
	public Vector3D(float xx, float yy, float zz) {
		set(xx, yy, zz);
	}

	/**
	 * Creates a vector by interpreting the passed point as a vector and copying
	 * its coordinates.
	 */
	public Vector3D(Point3D p) {
		set(p);
	}

	/**
	 * Creates a vector that points from point <code>from</code> to point
	 * <code>to</code>.
	 */
	public Vector3D(Point3D from, Point3D to) {
		set(from, to);
	}

	/**
	 * Sets this vector to point in the given direction. Returns a reference to
	 * this vector.
	 */
	public Vector3D set(float xx, float yy, float zz) {
		x = xx;
		y = yy;
		z = zz;
		return this;
	}

	/**
	 * Sets this vector to point in the same direction as the passed vector.
	 * Returns a reference to this vector.
	 */
	public Vector3D set(Vector3D v) {
		x = v.x;
		y = v.y;
		z = v.z;
		return this;
	}

	/**
	 * Interprets the passed point as a vector and sets this vector
	 * accordingly.. Same as <code>set(new Point3D(0,0,0), p)</code>. Returns
	 * a reference to this vector.
	 */
	public Vector3D set(Point3D p) {
		x = p.x;
		y = p.y;
		z = p.z;
		return this;
	}

	/**
	 * Sets this vector to point from <code>from</code> to <code>to</code>.
	 * Returns a reference to this vector.
	 */
	public Vector3D set(Point3D from, Point3D to) {
		x = to.x - from.x;
		y = to.y - from.y;
		z = to.z - from.z;
		return this;
	}

	/**
	 * Returns <code>true</code> if this is a null-vector (all entries are
	 * <code>0</code>), <code>false</code> otherwise.
	 */
	public boolean isZero() {
		return x == 0 && y == 0 && z == 0;
	}

	/**
	 * Returns the length of this vector.
	 */
	public float length() {
		return (float) Math.sqrt((double) (x * x + y * y + z * z));
	}

	/**
	 * Normalizes this vector. After normalization, this vector's
	 * <code>length()</code> equals <code>1</code>. Returns a reference to
	 * this vector.
	 */
	public Vector3D normalize() {
		float l = 1.0f / length();
		x *= l;
		y *= l;
		z *= l;
		return this;
	}

	/**
	 * Normalizes a copy of this vector and returns the copy.
	 */
	public Vector3D getNormalized() {
		Vector3D n = new Vector3D(this);
		return n.normalize();
	}

	/**
	 * Computes the cross product of this vector with the passed one and stores
	 * the result in this vector. Then returns a reference to this vector.
	 * <p>
	 * In a right handed coordinate system, where X, Y and Z denote the vectors
	 * of the axes, we have: <code>Z = X.cross(Y)</code> and
	 * <code>-Z - Y.cross(X)</code>.
	 */
	public Vector3D cross(Vector3D v) {
		float xx = y * v.z - z * v.y;
		float yy = z * v.x - x * v.z;
		float zz = x * v.y - y * v.x;
		x = xx;
		y = yy;
		z = zz;
		return this;
	}

	/**
	 * Computes the cross product of this vector with the passed one and returns
	 * the result.
	 */
	public Vector3D getCrossed(Vector3D v) {
		Vector3D c = new Vector3D(this);
		return c.cross(v);
	}

	/**
	 * Reverses the direction of this vector. Returns a reference to this
	 * vector.
	 */
	public Vector3D reverse() {
		x = -x;
		y = -y;
		z = -z;
		return this;
	}

	/**
	 * Reverses a copy of this vector and returns the copy.
	 */
	public Vector3D getReversed() {
		Vector3D v = new Vector3D(this);
		return v.reverse();
	}

	/**
	 * Scales this vector by the given scaling factor. Returns a reference to
	 * this vector.
	 */
	public Vector3D scale(float s) {
		x *= s;
		y *= s;
		z *= s;
		return this;
	}

	/**
	 * Scales a copy of this vector by the given scaling factor and returns the
	 * copy.
	 */
	public Vector3D getScaled(float s) {
		Vector3D v = new Vector3D(this);
		return v.scale(s);
	}

	/**
	 * Adds the given vector to this one. Returns a reference to this vector.
	 */
	public Vector3D add(Vector3D v) {
		x += v.x;
		y += v.y;
		z += v.z;
		return this;
	}

	/**
	 * Adds the given vector to a copy of this one and returns the copy.
	 */
	public Vector3D getAdded(Vector3D a) {
		Vector3D v = new Vector3D(this);
		return v.add(a);
	}

	/**
	 * Subtracts the given vector from this one. Returns a reference to this
	 * vector.
	 */
	public Vector3D subtract(Vector3D v) {
		x -= v.x;
		y -= v.y;
		z -= v.z;
		return this;
	}

	/**
	 * Subtracts the given vector from a copy of this one and returns the copy.
	 */
	public Vector3D getSubtracted(Vector3D a) {
		Vector3D v = new Vector3D(this);
		return v.subtract(a);
	}

	/**
	 * Computes the dot product of this vector and the passed one.
	 */
	public float dot(Vector3D v) {
		return x * v.x + y * v.y + z * v.z;
	}

	public String toString() {
		return "[" + String.valueOf(x) + " " + String.valueOf(y) + " "
				+ String.valueOf(z) + "]";
	}

	/**
	 * Determines the least dominant axis of this vector and returns its index.
	 * <p>
	 * The least dominant axis of a vector is the one with the smallest absolute
	 * value, e.g. for a vector <code>(2,-3,1)</code>, the least dominant
	 * axis is the z-axis.
	 * <p>
	 * The x-axis has index <code>0</code>, the y-axis has index
	 * <code>1</code> and the z-axis has index <code>2</code>.
	 */
	public int getMinAxis() {
		float xx = Math.abs(x);
		float yy = Math.abs(y);
		float zz = Math.abs(z);
		if (xx <= yy && xx <= zz)
			return 0;
		else if (yy <= xx && yy <= zz)
			return 1;
		else if (zz <= xx && zz <= yy)
			return 2;
		return 0; // program should never reach this point
	}

	/**
	 * Determines the most dominant axis of this vector and returns its index.
	 * <p>
	 * The most dominant axis of a vector is the one with the highest absolute
	 * value, e.g. for a vector <code>(2,-3,1)</code>, the most dominant axis
	 * is the y-axis.
	 * <p>
	 * The x-axis has index <code>0</code>, the y-axis has index
	 * <code>1</code> and the z-axis has index <code>2</code>.
	 */
	public int getMaxAxis() {
		float xx = Math.abs(x);
		float yy = Math.abs(y);
		float zz = Math.abs(z);
		if (xx >= yy && xx >= zz)
			return 0;
		else if (yy >= xx && yy >= zz)
			return 1;
		else if (zz >= xx && zz >= yy)
			return 2;
		return 0; // program should never reach this point
	}

	/**
	 * Returns a vector that is orthogonal to this one.
	 * <p>
	 * Note that there are infinitely many such vectors (and they all lie on a
	 * plane), so this method just returns an arbitrary one of them.
	 */
	public Vector3D getOrthogonal() {

		float current[] = { x, y, z, };
		float orthogonal[] = { 0, 0, 0 };
		int axis1 = getMaxAxis();
		int axis2 = (axis1 + 1) % 3;
		int axis3 = (axis1 + 2) % 3;

		// use projection of vector on axis1/axis2 plane to find the
		// perpendicular vector
		orthogonal[axis1] = -current[axis2];
		orthogonal[axis2] = current[axis1];
		orthogonal[axis3] = 0;

		return new Vector3D(orthogonal[0], orthogonal[1], orthogonal[2]);
	}

	/**
	 * Returns a <code>Point3D</code> with the same x,y,z entries as this
	 * vector.
	 */
	public Point3D getPoint3D() {
		return new Point3D(x, y, z);
	}
}
