/**
 * y25 - Visualizing Graphs in 2.5 Dimensions
 * 
 * Arbeitsbereich: Paralleles Rechnen Prof. Kaufmann
 * Studienarbeit:  Werner Jainek
 * Betreuung:      Nina Lehmann
 *
 */
package y25.io;

import java.awt.Color;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import y.base.Edge;
import y.base.EdgeCursor;
import y.base.Graph;
import y.base.Node;
import y.io.GMLIOHandler;
import y.io.gml.EdgeGraphicsParser;
import y.io.gml.EdgeObjectEncoder;
import y.io.gml.EdgeParser;
import y.io.gml.EdgeRealizerObjectEncoder;
import y.io.gml.EncoderFactory;
import y.io.gml.GMLEncoder;
import y.io.gml.GmlObjectEncoder;
import y.io.gml.GraphObjectEncoder;
import y.io.gml.GraphParser;
import y.io.gml.ItemParser;
import y.io.gml.NodeGraphicsParser;
import y.io.gml.NodeObjectEncoder;
import y.io.gml.NodeParser;
import y.io.gml.NodeRealizerObjectEncoder;
import y.io.gml.ObjectEncoder;
import y.io.gml.ParserFactory;
import y.io.gml.GMLTokenizer.Callback;
import y.view.EdgeRealizer;
import y.view.Graph2D;
import y.view.NodeRealizer;
import y25.base.GraphCursor;
import y25.base.LayerEdge;
import y25.base.LayerEdgeCursor;
import y25.graphics.BoundingBox;
import y25.graphics.Point3D;
import y25.view.EdgeRealizer25D;
import y25.view.Graph25D;
import y25.view.LayerEdgeRealizer25D;
import y25.view.LayerRealizer25D;
import y25.view.NodeRealizer25D;
import y25.view.realizer.PolyLineEdgeRealizer25D;
import y25.view.realizer.PolyLineLayerEdgeRealizer25D;
import y25.view.realizer.Realizer25DFactory;
import y25.view.realizer.ShapeNodeRealizer25D;
import y25.view.realizer.TransparentLayerRealizer25D;

/**
 * Reads and writes 2.5d graphs in the GML format from and to files.
 * 
 * @author Werner Jainek
 */
public class GMLIOHandler25D {

	// ////////////////////////////////////////////////////////////////////////////////////
	// 
	// W R I T I N G G R A P H S
	//
	// ////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Writes the passed 2.5d graph to the file given by the passed string.
	 */
	public void write(Graph25D graph25D, String name) throws IOException {

		// create the io handler
		GML25DEncoderFactory encoderFactory = new GML25DEncoderFactory(graph25D);
		GMLIOHandler.setEncoderFactory(encoderFactory);
		GMLIOHandler ioh = new GMLIOHandler();

		// create an output stream
		OutputStream out = new FileOutputStream(name, false);

		// check if we have at least one graph inside
		Graph2D g2d = graph25D.G() > 0 ? (Graph2D) graph25D.graphs().graph()
				: null;
		ioh.write(g2d, out);
	}

	private class GML25DEncoderFactory implements EncoderFactory {

		protected Graph25D graph25D;

		protected HashMap graphOrder; // used for layer edge encoding

		protected Graph2D dummyGraph;

		protected Edge dummyEdge;

		public GML25DEncoderFactory(Graph25D g25d) {
			graph25D = g25d;
			graphOrder = new HashMap();
			int i = 0;
			for (GraphCursor gc = g25d.graphs(); gc.ok(); gc.next()) {
				graphOrder.put(gc.graph(), new Integer(i));
				++i;
			}

			dummyGraph = new Graph2D(); // for default realizers
			dummyEdge = dummyGraph.createEdge(dummyGraph.createNode(),
					dummyGraph.createNode());
		}

		// ENCODER

		public ObjectEncoder createGMLEncoder() {
			Gml25DObjectEncoder gmlEncoder = new Gml25DObjectEncoder();
			gmlEncoder.setGraphEncoder(createGraphEncoder(gmlEncoder));
			gmlEncoder.setLayerEdgeEncoder(createLayerEdgeEncoder(gmlEncoder));
			return gmlEncoder;
		}

		public ObjectEncoder createLayerEdgeEncoder(ObjectEncoder gmlEncoder) {
			LayerEdgeRealizer25DObjectEncoder realizerEncoder = new LayerEdgeRealizer25DObjectEncoder();
			LayerEdge25DObjectEncoder layerEdgeEncoder = new LayerEdge25DObjectEncoder(
					realizerEncoder);
			return layerEdgeEncoder;
		}

		public ObjectEncoder createGraphEncoder(ObjectEncoder gmlEncoder) {
			GraphObjectEncoder graphEncoder = new Graph25DObjectEncoder();
			graphEncoder.setEdgeEncoder(createEdgeEncoder(graphEncoder));
			graphEncoder.setNodeEncoder(createNodeEncoder(graphEncoder));
			return graphEncoder;
		}

		public ObjectEncoder createNodeEncoder(ObjectEncoder graphEncoder) {
			ObjectEncoder realizerEncoder = new NodeRealizer25DObjectEncoder();
			ObjectEncoder nodeEncoder = new Node25DObjectEncoder(
					realizerEncoder, null);
			return nodeEncoder;
		}

		public ObjectEncoder createEdgeEncoder(ObjectEncoder graphEncoder) {
			ObjectEncoder realizerEncoder = new EdgeRealizer25DObjectEncoder();
			ObjectEncoder edgeEncoder = new Edge25DObjectEncoder(
					realizerEncoder, null);
			return edgeEncoder;
		}

		// GML ENCODER //ignores passed graph and writes the internal graph25d
		// to the output !
		public class Gml25DObjectEncoder extends GmlObjectEncoder {
			protected LayerEdge25DObjectEncoder layerEdgeEncoder;

			public void setLayerEdgeEncoder(ObjectEncoder lenc) {
				layerEdgeEncoder = (LayerEdge25DObjectEncoder) lenc;
			}

			public LayerEdge25DObjectEncoder getLayerEdgeEncoder() {
				return layerEdgeEncoder;
			}

			public void encode(Object item, GMLEncoder enc) throws IOException {
				enc.addAttribute("y25Version", "1.0");
				enc.beginSection("graph25d");

				// write all graphs
				for (GraphCursor gc = graph25D.graphs(); gc.ok(); gc.next()) {
					enc.beginSection("graph");
					getGraphEncoder().encode(gc.graph(), enc);
					enc.endSection();
				}

				// write all layer edges
				for (LayerEdgeCursor lec = graph25D.layerEdges(); lec.ok(); lec
						.next()) {
					enc.beginSection("layerEdge");
					getLayerEdgeEncoder().encode(lec.layerEdge(), enc);
					enc.endSection();
				}

				// write default node/edge/layeredge/layer realizers
				enc.beginSection("defaultNodeGraphics");
				new NodeRealizer25DObjectEncoder().encode(graph25D
						.getDefaultNodeRealizer25D(), enc);
				enc.endSection();

				enc.beginSection("defaultEdgeGraphics");
				EdgeRealizer er = (EdgeRealizer) graph25D
						.getDefaultEdgeRealizer25D();
				dummyGraph.setRealizer(dummyEdge, er);
				new EdgeRealizer25DObjectEncoder().encode(dummyGraph
						.getRealizer(dummyEdge), enc);
				enc.endSection();

				enc.beginSection("defaultLayerEdgeGraphics");
				new LayerEdgeRealizer25DObjectEncoder().encode(graph25D
						.getDefaultLayerEdgeRealizer25D(), enc);
				enc.endSection();

				enc.beginSection("defaultLayerGraphics");
				new LayerRealizer25DObjectEncoder().encode(graph25D
						.getDefaultLayerRealizer25D(), enc);
				enc.endSection();

				enc.endSection();
			}
		}

		// LAYER EDGE ENCODER
		// ////////////////////////////////////////////////////////

		public class LayerEdge25DObjectEncoder implements ObjectEncoder {
			protected LayerEdgeRealizer25DObjectEncoder layerEdgeRealizer25Encoder;

			public LayerEdge25DObjectEncoder(
					LayerEdgeRealizer25DObjectEncoder realizerEncoder) {
				layerEdgeRealizer25Encoder = realizerEncoder;
			}

			public void encode(Object item, GMLEncoder enc) throws IOException {
				LayerEdge le = (LayerEdge) item;
				LayerEdgeRealizer25D ler = (LayerEdgeRealizer25D) graph25D
						.getRealizer(le);

				Integer srcGraph = (Integer) graphOrder
						.get(le.getSourceGraph());
				Integer trgGraph = (Integer) graphOrder
						.get(le.getTargetGraph());

				enc.addAttribute("sourceGraph", srcGraph.intValue());
				enc.addAttribute("sourceNode", le.source().index());
				enc.addAttribute("targetGraph", trgGraph.intValue());
				enc.addAttribute("targetNode", le.target().index());

				if (layerEdgeRealizer25Encoder != null) {
					enc.beginSection("graphics");
					layerEdgeRealizer25Encoder.encode(ler, enc);
					enc.endSection();
				}
			}
		}

		public class LayerEdgeRealizer25DObjectEncoder implements ObjectEncoder {
			public void encode(Object item, GMLEncoder enc) throws IOException {
				LayerEdgeRealizer25D ler = (LayerEdgeRealizer25D) item;

				enc.addAttribute("radius", ler.getLineRadius());

				enc.addAttribute("color", GMLIOHandler.convertColor(ler
						.getLineColor()));

				// line display mode
				int ldm = ler.getLineDisplayMode();
				if (ldm == LayerEdgeRealizer25D.MODE_2D) {
					enc.addAttribute("lineDisplayMode", "2D");
				} else {
					enc.addAttribute("lineDisplayMode", "3D");
				}

				// arrows
				enc.addAttribute("sourceArrow", EdgeRealizerObjectEncoder
						.encodeArrowType(ler.getSourceArrow()));
				enc.addAttribute("targetArrow", EdgeRealizerObjectEncoder
						.encodeArrowType(ler.getTargetArrow()));
			}
		}

		// GRAPH ENCODER
		// ////////////////////////////////////////////////////////

		public class Graph25DObjectEncoder extends GraphObjectEncoder {
			public void encode(Object item, GMLEncoder enc) throws IOException {

				Graph2D graph = (Graph2D) item;
				LayerRealizer25D lr = graph25D.getRealizer(graph);

				// label of the layer
				enc.addAttribute("layerLabel", lr.getLabelText());

				// encode graph nodes and edges
				super.encode(graph, enc);

				// encode layer realizer information
				enc.beginSection("graphics");
				new LayerRealizer25DObjectEncoder().encode(lr, enc);
				enc.endSection();

			}
		}

		public class LayerRealizer25DObjectEncoder implements ObjectEncoder {
			public void encode(Object item, GMLEncoder enc) throws IOException {
				LayerRealizer25D lr = (LayerRealizer25D) item;

				BoundingBox bb = lr.getRealativeBoundingBox();
				enc.addAttribute("layerZ", lr.getZ());
				enc.addAttribute("layerBBCollapsed", bb.isCollapsed());
				if (!bb.isCollapsed()) {
					enc.addAttribute("layerBBMinX", bb.getMin().x);
					enc.addAttribute("layerBBMinY", bb.getMin().y);
					enc.addAttribute("layerBBMinZ", bb.getMin().z);
					enc.addAttribute("layerBBMaxX", bb.getMax().x);
					enc.addAttribute("layerBBMaxY", bb.getMax().y);
					enc.addAttribute("layerBBMaxZ", bb.getMax().z);
				}
			}
		}

		// NODE ENCODER////////////////////////////////////////////////////////

		public class Node25DObjectEncoder extends NodeObjectEncoder {
			public Node25DObjectEncoder(ObjectEncoder realizerEncoder,
					ObjectEncoder labelEncoder) {
				super(realizerEncoder, labelEncoder);
			}

			public void encode(Object item, GMLEncoder enc) throws IOException {
				Node node = (Node) item;
				enc.addAttribute("nodeID", graph25D.getNodeID(node));
				enc.addAttribute("parentID", graph25D.getParentID(node));
				super.encode(item, enc);
			}
		}

		public class NodeRealizer25DObjectEncoder extends
				NodeRealizerObjectEncoder {

			public void encode(Object item, GMLEncoder enc) throws IOException {
				super.encode(item, enc);

				NodeRealizer25D nr = (NodeRealizer25D) item;
				enc.addAttribute("d", nr.getDepth());
			}
		}

		// EDGE ENCODER ////////////////////////////////////////////////////////

		public class Edge25DObjectEncoder extends EdgeObjectEncoder {
			public Edge25DObjectEncoder(ObjectEncoder realizerEncoder,
					ObjectEncoder labelEncoder) {
				super(realizerEncoder, labelEncoder);
			}

			public void encode(Object item, GMLEncoder enc) throws IOException {
				Edge edge = (Edge) item;
				enc.addAttribute("edgeID", graph25D.getEdgeID(edge));
				super.encode(item, enc);
			}
		}

		public class EdgeRealizer25DObjectEncoder extends
				EdgeRealizerObjectEncoder {

			public void encode(Object item, GMLEncoder enc) throws IOException {
				super.encode(item, enc);

				EdgeRealizer25D er = (EdgeRealizer25D) item;

				// line display mode
				int ldm = er.getLineDisplayMode();
				if (ldm == EdgeRealizer25D.MODE_2D) {
					enc.addAttribute("lineDisplayMode", "2D");
				} else {
					enc.addAttribute("lineDisplayMode", "3D");
				}
			}
		}

	}

	// ////////////////////////////////////////////////////////////////////////////////////
	// 
	// R E A D I N G G R A P H S
	//
	// ////////////////////////////////////////////////////////////////////////////////////

	/**
	 * Reads a 2.5 graph from the file with the passed name into the passed
	 * <code>Graph25D</code>.
	 */
	public void read(Graph25D graph25D, String name) throws IOException {

		// clear the passed graph25D
		graph25D.clear();

		// create the io handler for one graph
		GML25DParserFactory parserFactory = new GML25DParserFactory(graph25D);
		GMLIOHandler.setParserFactory(parserFactory);
		GMLIOHandler ioh = new GMLIOHandler();

		// read into one graph - this graph is entirely ignored, input is beeing
		// written into graph25D
		Graph2D g = new Graph2D();
		ioh.read(g, name);
	}

	private class GML25DParserFactory implements ParserFactory {

		protected Graph25D graph25D;

		protected Graph2D dummyGraph;

		protected Edge dummyEdge;

		public GML25DParserFactory(Graph25D g25d) {
			graph25D = g25d;
			dummyGraph = new Graph2D(); // for default realizers
			dummyEdge = dummyGraph.createEdge(dummyGraph.createNode(),
					dummyGraph.createNode());
		}

		// PARSER

		public Callback createGMLParser(Graph2D graph) {
			System.out.println("A");
			ItemParser gmlParser = new Gml25DParser();
			return gmlParser;
		}

		public Callback createGraphParser(Graph2D graph, Callback gmlParser) {
			// System.out.println("B");
			// ItemParser graphParser = new Graph2DParser(graph);
			// graphParser.addChild("node", (ItemParser) createNodeParser(graph,
			// graphParser));
			// graphParser.addChild("edge", (ItemParser) createEdgeParser(graph,
			// graphParser));
			return new ItemParser();
		}

		public Callback createNodeParser(Graph2D graph, Callback graphParser) {
			// System.out.println("C");
			// GraphParser gp = (GraphParser) graphParser;
			// ItemParser parser = new Node25DParser(graph, gp.getId2Node());
			return new ItemParser();
		}

		public Callback createEdgeParser(Graph2D graph, Callback graphParser) {
			// System.out.println("D");
			// GraphParser gp = (GraphParser) graphParser;
			// ItemParser parser = new EdgeParser(graph, gp.getId2Node(),
			// gp.getId2Edge());
			return new ItemParser();
		}

		// GML PARSER
		public class Gml25DParser extends ItemParser {
			public Gml25DParser() {
				super();
				addChild("graph25d", new Graph25DParser());
			}
		}

		// GRAPH 25D PARSER
		public class Graph25DParser extends ItemParser {

			NodeGraphics25DParser defaultNodeGraphicsParser;

			EdgeGraphicsParser defaultEdgeGraphicsParser;

			LayerEdgeGraphics25DParser defaultLayerEdgeGraphicsParser;

			LayerGraphics25DParser defaultLayerGraphicsParser;

			public Graph25DParser() {

				try {
					defaultNodeGraphicsParser = new NodeGraphics25DParser(
							new URL("http://www.url.com"));
				} catch (MalformedURLException e) {
				}
				defaultEdgeGraphicsParser = new EdgeGraphicsParser();
				defaultLayerEdgeGraphicsParser = new LayerEdgeGraphics25DParser();
				defaultLayerGraphicsParser = new LayerGraphics25DParser();

				addChild("graph", new Graph2DParser(dummyGraph));
				addChild("layerEdge", new LayerEdge25DParser());
				addChild("defaultNodeGraphics", defaultNodeGraphicsParser);
				addChild("defaultEdgeGraphics", defaultEdgeGraphicsParser);
				addChild("defaultLayerEdgeGraphics",
						defaultLayerEdgeGraphicsParser);
				addChild("defaultLayerGraphics", defaultLayerGraphicsParser);
			}

			public void begin() {
				super.begin();
				// System.out.println("G25");

			}

			public void end() {
				super.end();

				// default node realizer
				graph25D.setDefaultNodeRealizer25D(defaultNodeGraphicsParser
						.getNodeRealizer());

				// default edge realizer
				Realizer25DFactory f = new Realizer25DFactory(
						new ShapeNodeRealizer25D(),
						new PolyLineEdgeRealizer25D());
				graph25D.setDefaultEdgeRealizer25D(f
						.copyEdgeRealizer(defaultEdgeGraphicsParser
								.getEdgeRealizer()));

				// default layer edge realizer
				LayerEdgeRealizer25D ler = (LayerEdgeRealizer25D) defaultLayerEdgeGraphicsParser
						.getItem();
				graph25D.setDefaultLayerEdgeRealizer25D(ler);

				// default layer realizer
				LayerRealizer25D lr = (LayerRealizer25D) defaultLayerGraphicsParser
						.getItem();
				graph25D.setDefaultLayerRealizer25D(lr);

				// System.out.println("/G25");
			}
		}

		// GRAPH PARSER
		public class Graph2DParser extends GraphParser {
			Graph2D currentGraph;

			LayerRealizer25D layerRealizer;

			LayerGraphics25DParser layerGraphicsParser;

			public Graph2DParser(Graph2D dummy) {
				super(dummy);
				layerRealizer = new TransparentLayerRealizer25D();
				layerGraphicsParser = new LayerGraphics25DParser();

				addChild("graphics", layerGraphicsParser);
			}

			public void childFinished(ItemParser p) {
				if (p == layerGraphicsParser) {
					layerRealizer = (LayerRealizer25D) layerGraphicsParser
							.getItem();
				}
			}

			public void begin() {
				// create a new graph inside graph25d
				currentGraph = (Graph2D) graph25D.createGraph();
				graph = currentGraph;
				id2Node.clear();
				id2Edge.clear();
				addChild("node", new Node25DParser(currentGraph, id2Node));
				addChild("edge", new Edge25DParser(currentGraph, id2Node,
						id2Edge));
				super.begin();
				// System.out.println(" G");
			}

			public void end() {
				super.end();

				// change all edge realizers to 25d version
				Realizer25DFactory f = new Realizer25DFactory(
						new ShapeNodeRealizer25D(),
						new PolyLineEdgeRealizer25D());
				for (EdgeCursor ec = graph.edges(); ec.ok(); ec.next()) {
					Edge e = ec.edge();
					graph.setRealizer(e, f.copyEdgeRealizer(graph
							.getRealizer(e)));
				}

				// set label of this layer
				layerRealizer.setLabelText(getString("layerLabel"));
				graph25D.setRealizer(currentGraph, layerRealizer);

				// System.out.println(" /G");
			}

		}

		public class LayerGraphics25DParser extends ItemParser {
			LayerRealizer25D lr;

			public LayerGraphics25DParser() {
				super();
			}

			public void begin() {
				lr = new TransparentLayerRealizer25D();
				super.begin();
			}

			public void end() {
				lr.setZ(getDouble("layerZ", 0));

				if (getBoolean("layerBBCollapsed")) {
					lr.setRelativeBoundingBox(new BoundingBox());
				} else {
					lr.setRelativeBoundingBox(new BoundingBox(new Point3D(
							(float) getDouble("layerBBMinX"),
							(float) getDouble("layerBBMinY"),
							(float) getDouble("layerBBMinZ")), new Point3D(
							(float) getDouble("layerBBMaxX"),
							(float) getDouble("layerBBMaxY"),
							(float) getDouble("layerBBMaxZ"))));
				}
				super.end();
			}

			public Object getItem() {
				return lr;
			}
		}

		// NODE PARSER

		public class Node25DParser extends NodeParser {
			Graph2D currentGraph;

			public Node25DParser(Graph2D graph, Map id2Node) {

				super(graph, id2Node);
				currentGraph = graph;

				// default node realizer
				nr = new ShapeNodeRealizer25D();

				try {
					nodeGraphics = new NodeGraphics25DParser(new URL(
							"http://www.url.com"));
				} catch (MalformedURLException e) {
				}

				addChild("graphics", nodeGraphics);
			}

			public void begin() {
				super.begin();
				// System.out.println(" N");
			}

			public void end() {
				super.end();
				graph25D
						.setNodeID(currentGraph.lastNode(), getInt("nodeID", 0));
				graph25D.setParentID(currentGraph.lastNode(), getInt(
						"parentID", 0));
				// System.out.println(" /N");
			}
		}

		public class NodeGraphics25DParser extends NodeGraphicsParser {

			public NodeGraphics25DParser(URL contextURL) {
				super(contextURL);
			}

			public NodeRealizer getNodeRealizer() {
				NodeRealizer nr = super.getNodeRealizer();

				// convert realizer to 25D verison
				Realizer25DFactory f = new Realizer25DFactory(
						new ShapeNodeRealizer25D(),
						new PolyLineEdgeRealizer25D());
				NodeRealizer25D copy = (NodeRealizer25D) f.copyNodeRealizer(nr);

				// now add the 25D information
				copy.setDepth(getDouble("d", copy.getDepth()));

				return (NodeRealizer) copy;
			}

		}

		// EDGE PARSER

		public class Edge25DParser extends EdgeParser {
			Graph2D currentGraph;

			public Edge25DParser(Graph2D graph, Map id2Node, Map id2Edge) {
				super(graph, id2Node, id2Edge);
				currentGraph = graph;
			}

			public void begin() {
				super.begin();
				// System.out.println(" E");
			}

			public void end() {
				super.end();
				graph25D
						.setEdgeID(currentGraph.lastEdge(), getInt("edgeID", 0));
				// System.out.println(" /E");
			}
		}

		// LAYER EDGE PARSER
		public class LayerEdge25DParser extends ItemParser {
			private LayerEdgeRealizer25D layerEdgeRealizer;

			private ItemParser layerEdgeGraphicsParser;

			public LayerEdge25DParser() {
				super();
				layerEdgeRealizer = new PolyLineLayerEdgeRealizer25D();
				layerEdgeGraphicsParser = new LayerEdgeGraphics25DParser();
				addChild("graphics", layerEdgeGraphicsParser);
			}

			public void childFinished(ItemParser p) {
				if (p == layerEdgeGraphicsParser) {
					layerEdgeRealizer = (LayerEdgeRealizer25D) layerEdgeGraphicsParser
							.getItem();
				}
			}

			public void begin() {
				super.begin();
				// System.out.println(" LE");
			}

			public void end() {
				super.end();
				Graph graphs[] = graph25D.getGraphArray();
				Node srcNodes[] = graphs[getInt("sourceGraph")].getNodeArray();
				Node trgNodes[] = graphs[getInt("targetGraph")].getNodeArray();
				LayerEdge le = graph25D.createLayerEdge(
						srcNodes[getInt("sourceNode")],
						trgNodes[getInt("targetNode")]);
				graph25D.setRealizer(le, layerEdgeRealizer);
				// System.out.println(" /LE");
			}
		}

		public class LayerEdgeGraphics25DParser extends ItemParser {
			LayerEdgeRealizer25D ler;

			public LayerEdgeGraphics25DParser() {
				super();
			}

			public void begin() {
				ler = new PolyLineLayerEdgeRealizer25D();
				super.begin();
			}

			public void end() {
				ler.setLineColor(Color.decode(getString("color")));
				ler.setLineRadius(getDouble("radius"));
				ler.setSourceArrow(EdgeGraphicsParser
						.decodeArrow(getString("sourceArrow")));
				ler.setTargetArrow(EdgeGraphicsParser
						.decodeArrow(getString("targetArrow")));

				// line display mode
				String lineDisplayMode = getString("lineDisplayMode");
				if (lineDisplayMode == "2D") {
					ler.setLineDisplayMode(LayerEdgeRealizer25D.MODE_2D);
				} else {
					ler.setLineDisplayMode(LayerEdgeRealizer25D.MODE_3D);
				}

				super.end();
			}

			public Object getItem() {
				return ler;
			}
		}

	}

}
