/**
 * y25 - Visualizing Graphs in 2.5 Dimensions
 * 
 * Arbeitsbereich: Paralleles Rechnen Prof. Kaufmann
 * Studienarbeit:  Werner Jainek
 * Betreuung:      Nina Lehmann
 *
 */
package y25.layout;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;

import y.base.Edge;
import y.base.EdgeCursor;
import y.base.EdgeMap;
import y.base.Node;
import y.base.NodeCursor;
import y.base.NodeMap;
import y.geom.YPoint;
import y.layout.EdgeLayout;
import y.layout.LayoutGraph;
import y.layout.Layouter;
import y.layout.NodeLayout;
import y25.base.GraphCursor;
import y25.base.LayeredGraph;
import y25.view.Graph25D;

/**
 * First flattens a <code>LayeredLayoutGraph</code>, then applies a 2D
 * layouter to it, then unflattens it.
 * <p>
 * The flattening and unflattening of the layered layout graph is done by a
 * layout stage that is stored inside this class. This flattening/unflattening
 * can be customized by the user.
 * <p>
 * If no flattening/unflattening stage is provided, a canonic flattening stage
 * is used. This stage flattens all layers by combining the nodes and edges that
 * have the same id. The resulting flattened graph has all node and edge id's
 * that occur across all layers of the original layered layout graph. The
 * flattened graph is then passed to the 2D layouter. After it has finished
 * layouting, the unflattening stage simply copies all node and edge layout
 * information to all layers of the original layered layout graph.
 * 
 * @author Werner Jainek
 */
public class FlatLayouter implements Layouter25D {

	/**
	 * The composite layouter that does the flatteing, layout, unflattening.
	 */
	protected CompositeLayouter25D layouter;

	/**
	 * Creates a FlatLayouter with the canonic flattening/unflattening stage.
	 */
	public FlatLayouter(Layouter coreLayouter) {
		layouter = new CompositeLayouter25D(new CanonicFlatteningStage(),
				new Layouter25DAdapter(coreLayouter));
	}

	/**
	 * Creates a FlatLayouter with a user supplied flattening/unflattening
	 * stage.
	 */
	public FlatLayouter(LayoutStage25D flattener, Layouter coreLayouter) {
		layouter = new CompositeLayouter25D(flattener, new Layouter25DAdapter(
				coreLayouter));
	}

	public boolean canLayout(LayeredLayoutGraph llg) {
		return layouter.canLayout(llg);
	}

	public void doLayout(LayeredLayoutGraph llg) {
		layouter.doLayout(llg);
	}

	/**
	 * The canonic flatteing/unflattening stage used by FlatLayouter.
	 * 
	 * @author Werner Jainek
	 */
	protected class CanonicFlatteningStage extends AbstractLayoutStage25D {

		public boolean canLayout(LayeredLayoutGraph llg) {
			// flattening is always possible
			return canLayoutCore(llg);
		}

		public void doLayout(LayeredLayoutGraph llg) {

			// TODO: change this to LayeredLayoutGraph once
			// DefaultLayeredLayoutGraph is implemented !
			// create layered layout graph that will hold the flattened graph
			LayeredLayoutGraph flat25d = new Graph25D();
			LayoutGraph flat = (LayoutGraph) flat25d.createGraph();
			NodeMap flatNodeIDs = (NodeMap) flat
					.getDataProvider(LayeredGraph.NODE_ID_DPKEY);

			// ------- 1st FLATTEN THE GRAPH -------

			// store already encountered id's
			HashMap usedNodeIDs = new HashMap();
			HashMap usedEdgeIDs = new HashMap();

			// iterate through all graphs
			for (GraphCursor gc = llg.graphs(); gc.ok(); gc.next()) {

				LayoutGraph graph = (LayoutGraph) gc.graph();
				NodeMap nodeIDs = (NodeMap) graph
						.getDataProvider(LayeredGraph.NODE_ID_DPKEY);
				EdgeMap edgeIDs = (EdgeMap) graph
						.getDataProvider(LayeredGraph.EDGE_ID_DPKEY);

				// iterate through all nodes of current graph
				for (NodeCursor nc = graph.nodes(); nc.ok(); nc.next()) {

					// get the current node
					Node node = nc.node();

					// get the nodes id
					Integer nodeID = new Integer(nodeIDs.getInt(node));

					// only if we havent encountered that id yet ...
					if (!usedNodeIDs.containsKey(nodeID)) {

						// ... do we add this node to the flattened graph,
						Node flatNode = flat.createNode();
						flatNodeIDs.setInt(flatNode, nodeID.intValue());

						// ... copy all of the nodes properties
						NodeLayout nodeLayout = graph.getLayout(node);
						NodeLayout flatNodeLayout = flat.getLayout(flatNode);
						copyNodeLayout25D(nodeLayout, flatNodeLayout);

						// ... and record that we encountered this id
						usedNodeIDs.put(nodeID, flatNode);

					}

				}

				// iterate through all edges of the current graph
				for (EdgeCursor ec = graph.edges(); ec.ok(); ec.next()) {

					// get the current Edge
					Edge edge = ec.edge();

					// get the edge's id
					Integer edgeID = new Integer(edgeIDs.getInt(edge));

					// only if we haven't encountered that id yet ...
					if (!usedEdgeIDs.containsKey(edgeID)) {

						// ... do we get the id's of both nodes that edge
						// connects
						Integer sourceNodeID = new Integer(nodeIDs.getInt(edge
								.source()));
						Integer targetNodeID = new Integer(nodeIDs.getInt(edge
								.target()));

						// ... add this edge to the flattened graph
						Node flatNodeSource = getNodeByID(usedNodeIDs,
								sourceNodeID);
						Node flatNodeTarget = getNodeByID(usedNodeIDs,
								targetNodeID);
						Edge flatEdge = flat.createEdge(flatNodeSource,
								flatNodeTarget);

						// ... copy all of the edges properties
						EdgeLayout edgeLayout = graph.getLayout(edge);
						EdgeLayout flatEdgeLayout = flat.getLayout(flatEdge);
						copyEdgeLayout25D(edgeLayout, flatEdgeLayout);

						// ... and record that we encountered this edge id
						usedEdgeIDs.put(edgeID, flatEdge);
					}
				} // edges

			} // graphs

			// ------- 2nd LAYOUT FLAT GRAPH -------
			doLayoutCore(flat25d);

			// ------- 3rd UNFLATTEN GRAPH -------

			// iterate through all graphs
			for (GraphCursor gc = llg.graphs(); gc.ok(); gc.next()) {

				LayoutGraph graph = (LayoutGraph) gc.graph();
				NodeMap nodeIDs = (NodeMap) graph
						.getDataProvider(LayeredGraph.NODE_ID_DPKEY);
				EdgeMap edgeIDs = (EdgeMap) graph
						.getDataProvider(LayeredGraph.EDGE_ID_DPKEY);

				// iterate through all nodes of current graph
				for (NodeCursor nc = graph.nodes(); nc.ok(); nc.next()) {

					// get the current node
					Node node = nc.node();

					// get the nodes id
					Integer nodeID = new Integer(nodeIDs.getInt(node));

					// get the flatNode
					Node flatNode = getNodeByID(usedNodeIDs, nodeID);

					// ... copy all of the flat nodes properties
					NodeLayout nodeLayout = graph.getLayout(node);
					NodeLayout flatNodeLayout = flat.getLayout(flatNode);
					copyNodeLayout25D(flatNodeLayout, nodeLayout);
				}

				// iterate through all edges of the current graph
				for (EdgeCursor ec = graph.edges(); ec.ok(); ec.next()) {

					// get the current Edge
					Edge edge = ec.edge();

					// get the edge's id
					Integer edgeID = new Integer(edgeIDs.getInt(edge));

					// get the flat edge
					Edge flatEdge = getEdgeByID(usedEdgeIDs, edgeID);

					// ... copy all of the edges properties
					EdgeLayout edgeLayout = graph.getLayout(edge);
					EdgeLayout flatEdgeLayout = flat.getLayout(flatEdge);
					copyEdgeLayout25D(flatEdgeLayout, edgeLayout);

				} // edges

			} // graphs

		}

		protected Node getNodeByID(HashMap usedNodeIDs, Integer nodeID) {

			Set keys = usedNodeIDs.keySet();

			Iterator it = keys.iterator();
			while (it.hasNext()) {
				Integer i = (Integer) it.next();
				if (i.compareTo(nodeID) == 0) {
					return (Node) usedNodeIDs.get(i);
				}
			}
			return null; // should never happen !
		}

		protected Edge getEdgeByID(HashMap usedEdgeIDs, Integer edgeID) {

			Set keys = usedEdgeIDs.keySet();

			Iterator it = keys.iterator();
			while (it.hasNext()) {
				Integer i = (Integer) it.next();
				if (i.compareTo(edgeID) == 0) {
					return (Edge) usedEdgeIDs.get(i);
				}
			}
			return null; // should never happen !
		}

		protected void copyNodeLayout25D(NodeLayout src, NodeLayout trg) {

			// TODO: this should go to some internal method of NodeLayout25D !
			// or the layout should be copied when the new node is created, not
			// afterwards
			// (this can be done after DefaultLayeredLayoutGraph is implemented)

			// Node Layout Properties
			trg.setLocation(src.getX(), src.getY());
			trg.setSize(src.getWidth(), src.getHeight());

			// Node Layout 25D Properties
			NodeLayout25D src25 = (NodeLayout25D) src;
			NodeLayout25D trg25 = (NodeLayout25D) trg;

			src25.setDepth(trg25.getDepth());
		}

		protected void copyEdgeLayout25D(EdgeLayout src, EdgeLayout trg) {

			// Edge Layout Properties
			trg.setSourcePoint(src.getSourcePoint());
			trg.setTargetPoint(src.getTargetPoint());

			trg.clearPoints();
			for (int i = 0; i < src.pointCount(); ++i) {
				YPoint p = src.getPoint(i);
				trg.addPoint(p.x, p.y);
			}

			// Edge Layout 25D Properties
			// none yet
		}

	}

}
