/**
 * y25 - Visualizing Graphs in 2.5 Dimensions
 * 
 * Arbeitsbereich: Paralleles Rechnen Prof. Kaufmann
 * Studienarbeit:  Werner Jainek
 * Betreuung:      Nina Lehmann
 *
 */
package y25.layout;

import java.util.Comparator;

import y.base.EdgeCursor;
import y.base.Graph;
import y.base.Node;
import y.base.NodeCursor;
import y.layout.LayoutGraph;
import y25.base.GraphCursor;
import y25.base.GraphMap;
import y25.base.LayerEdge;
import y25.base.LayerEdgeCursor;
import y25.base.LayerEdgeMap;
import y25.base.LayeredGraph;
import y25.graphics.BoundingBox;
import y25.graphics.Point3D;
import y25.graphics.Vector3D;

/**
 * Extends <code>LayeredGraph</code> with layout information.
 * <p>
 * "Layout information" means all the information that is neccessary to arrange
 * elements of a <code>LayeredLayoutGraph</code> in three-dimensional space
 * (with an X,Y,Z coordinate system). There are some rules for the arrangement
 * of those entities in the three-dimensional space:
 * <ul>
 * <li>General
 * <ul>
 * <li>The (X,Y,Z) coordinate system is a left-handed coordinate system.
 * (Imagine a piece of paper lying on a desk in front of you. If the X-Axis goes
 * from left to right and the Y-Axis goes from top to bottom, then the Z-Axis
 * would stick out of the paper pointing upwards.)</li>
 * </ul>
 * </li>
 * <li>Layers
 * <ul>
 * <li>Every layer lies on a plane that is perpendicular to the Z-Axis. (In the
 * example before, the piece of paper on the desk would be a layer crossing the
 * Z-Axis at <code>0</code>.)</li>
 * <li>A layer has no depth / thickness - conceptually, it is a 2D-Subspace of
 * the 3D-Space the <code>LayeredLayoutGraph</code> resides in.</li>
 * <li>Inside each 2D-Subspace, a layer's extent is limited to a rectangular
 * area whose sides are perpendicular to the X- and Y-Axes respectively.</li>
 * </ul>
 * </li>
 * <li>LayerEdges
 * <ul>
 * <li>LayerEdges must start and end at nodes inside the
 * <code>LayeredLayoutGraph</code>, therefore, their endpoints are determined
 * by the layout information of those nodes. Other than that, there are no
 * constraints for the layout of layer edges.</li>
 * </ul>
 * </li>
 * <li>Nodes and Edges
 * <ul>
 * <li>All nodes and edges belong to a specific layer. Their Z-Coordinate is
 * fully determined by that information - they simply have the same Z-Coordinate
 * as the layer they belong to.</li>
 * <li>Nodes have a depth that determines their extent along the Z-Axis.</li>
 * <li>Apart from that, nodes and edges have all the layout information from
 * the <code>y.layout</code> package.
 * </ul>
 * </li>
 * </ul>
 * 
 * <p>
 * For every entity (layer, layer edge, node, edge) inside this
 * <code>LayeredLayoutGraph</code>, layout information is stored inside
 * objects that have the interfaces <code>LayerLayout25D</code>,
 * <code>LayerEdgeLayout25D</code>, <code>NodeLayout25D</code> and
 * <code>EdgeLayout25D</code> respectively. Wehereas layer and layer edge
 * information is stored directly inside this <code>LayeredLayoutGraph</code>,
 * node and edge layout information is stored inside each layer (graph).
 * <p>
 * Note that this is an abstract class and all subclasses need to implement at
 * least the following methods form <code>LayeredGraph</code>
 * <ul>
 * <li>createGraph</li>
 * <li>createGraph(int, Graph)</li>
 * <li>removeGraph</li>
 * <li>createLayerEdge</li>
 * <li>removeLayerEdge</li>
 * </ul>
 * and make sure that appropriate layouts are created / removed for every graph,
 * layer edge, node or edge that is created / removed. While the information for
 * layers (graphs) and layer edges is stored inside the
 * <code>LayeredLayoutGraph</code> directly, layout information for nodes and
 * edges are stored inside each layer (graph). As a result of this convention,
 * there is a natural relationship between the LayeredGraph-Hierarchy in the y25
 * package and the Graph-Hierarchy inside the y package: LayeredGraph stores
 * Graphs for each layer, LayeredLayoutGraph stores LayoutGraphs for each layer
 * and Graph25D stores Graph2Ds for each layer.
 * 
 * @author Werner Jainek
 * 
 * TODO: Right now, the bounding box is not recalculated when layout information
 * for nodes/edges changes. Probably a graph listener would be necessary to do
 * just that....
 */
public abstract class LayeredLayoutGraph extends LayeredGraph {

	/**
	 * Stores the layout information for each layer (graph).
	 */
	protected GraphMap layerLayouts;

	/**
	 * Stores the layout information for each layer edge.
	 */
	protected LayerEdgeMap layerEdgeLayouts;

	/**
	 * Stores the last computed bounding box for this
	 * <code>LayeredLayoutGraph</code>.
	 */
	protected BoundingBox boundingBox;

	// CONSTRUCTOR / RESETTING --------------------------------------------

	/**
	 * Creates an empty <code>LayeredLayoutGraph</code>.
	 */
	public LayeredLayoutGraph() {
		graphType = LayoutGraph.class;
		layerLayouts = createGraphMap();
		layerEdgeLayouts = createLayerEdgeMap();
		boundingBox = new BoundingBox();
	}

	/**
	 * Resets the <code>LayeredLayoutGraph</code> to its initial state.
	 */
	public void clear() {
		layerLayouts = createGraphMap();
		layerEdgeLayouts = createLayerEdgeMap();
		boundingBox = new BoundingBox();
		super.clear();
	}

	/**
	 * Removes all layer edges present in this <code>LayeredLayoutGraph</code>.
	 */
	public void clearLayerEdges() {
		layerEdgeLayouts = createLayerEdgeMap();
		super.clearLayerEdges();
		recalculateBoundingBox();
	}

	// GRAPHS --------------------------------------------

	// createGraph()
	// createGraph(int where, Graph reference)
	// duplicateGraph(Graph original)
	// importGraph(Graph graph)

	/**
	 * Removes the passed <code>LayoutGraph</code> from this
	 * <code>LayeredLayoutGraph</code>.
	 * <p>
	 * Additionally to the actions described in
	 * {@link y25.base.LayeredGraph#removeGraph(Graph) LayeredGraph.removeGraph}
	 * this method also removes the layer layout associated with the removed
	 * graph.
	 * 
	 * @param graph
	 *            The graph that will be removed.
	 * 
	 */
	public void removeGraph(Graph graph) {
		layerLayouts.remove(graph);
		super.removeGraph(graph);

		recalculateBoundingBox();
	}

	// LAYER EDGES ---------------------------------------

	// createLayerEdge (Graph src, Graph trg)

	/**
	 * Removes the passed layer edge from this LayeredLayoutGraph.
	 * <p>
	 * Additionally to the actions described in
	 * {@link y25.base.LayeredGraph#removeLayerEdge(LayerEdge) LayeredGraph.removeLayerEdge}
	 * this method also removes the layer edge layout associated with the
	 * removed layer edge.
	 * 
	 * @param layerEdge
	 *            The layer edge that will be removed.
	 */
	public void removeLayerEdge(LayerEdge layerEdge) {
		layerEdgeLayouts.remove(layerEdge);
		super.removeLayerEdge(layerEdge);

		recalculateBoundingBox();
	}

	// ITERATION / ORDER --------------------------------------------

	/**
	 * Sorts the layers (graphs) inside this <code>LayeredLayoutGraph</code>
	 * with respect to their z-position. Layers with low z-values will go before
	 * layers with higher z-values.
	 */
	public void sortGraphsByZ() {
		sortGraphs(new Comparator() {
			public int compare(Object g1, Object g2) {
				double z1 = getZ((Graph) g1);
				double z2 = getZ((Graph) g2);
				if (z1 == z2)
					return 0;
				else
					return (z1 < z2) ? -1 : 1;
			}
		});
	}

	// LAYOUT --------------------------------------------

	/**
	 * Returns a bounding box that encompasses all layers, layer edges, nodes
	 * and edges from this <code>LayeredLayoutGraph</code>.
	 */
	public BoundingBox getBoundingBox() {
		return new BoundingBox(boundingBox);
	}

	/**
	 * Erases the previously computed bounding box or this graph and recomputes
	 * it.
	 */
	public void recalculateBoundingBox() {
		// TODO: include labels in bounding boxes

		boundingBox.setCollapsed();

		// iterate through all graphs
		LayoutGraph g;
		for (GraphCursor gc = graphs(); gc.ok(); gc.next()) {

			g = (LayoutGraph) gc.graph();
			Vector3D offset = new Vector3D(0, 0, (float) getZ(g));

			// include bounding boxes of layer realizers
			{
				BoundingBox bb = ((LayerLayout25D) getLayout(g))
						.getRealativeBoundingBox();
				boundingBox.include(bb.translate(offset));
			}

			// include bounding boxes of nodes
			for (NodeCursor nc = g.nodes(); nc.ok(); nc.next()) {
				BoundingBox bb = ((NodeLayout25D) g.getLayout(nc.node()))
						.getRelativeBoundingBox();
				boundingBox.include(bb.translate(offset));
			}

			// include bounding boxes of edges
			for (EdgeCursor ec = g.edges(); ec.ok(); ec.next()) {
				BoundingBox bb = ((EdgeLayout25D) g.getLayout(ec.edge()))
						.getRelativeBoundingBox();
				boundingBox.include(bb.translate(offset));
			}
		}

		// include all bounding boxes of layer edges
		for (LayerEdgeCursor lec = layerEdges(); lec.ok(); lec.next()) {
			BoundingBox bb = ((LayerEdgeLayout25D) getLayout(lec.layerEdge()))
					.getBoundingBox();
			boundingBox.include(bb);
		}

		// conservative calculation, just add 20 to be sure everything is inside
		boundingBox.addBorder(20.0f);
	}

	/**
	 * Returns the layout associated with the given graph. The graph must belong
	 * to this <code>LayeredLayoutGraph</code>.
	 */
	public LayerLayout25D getLayout(Graph graph) {
		return (LayerLayout25D) layerLayouts.get(graph);
	}

	/**
	 * Returns the layout associated with the given layer edge. The layer edge
	 * must belong to this <code>LayeredLayoutGraph</code>.
	 */
	public LayerEdgeLayout25D getLayout(LayerEdge layerEdge) {
		return (LayerEdgeLayout25D) layerEdgeLayouts.get(layerEdge);
	}

	// LAYOUT SHORTCUTS

	/**
	 * Returns the position of the given node in three-dimensional space. X- and
	 * y-coordinates are extracted from the <code>NodeLayout25D</code> of the
	 * node, the z-coordinate is extracted from the LayerLayout25D of the layer
	 * the node belongs to.
	 */
	public Point3D getPoint3D(Node n) {

		LayoutGraph g = (LayoutGraph) n.getGraph();
		float nodeX = (float) g.getCenterX(n);
		float nodeY = (float) g.getCenterY(n);
		float nodeZ = (float) getZ(n.getGraph());
		return new Point3D(nodeX, nodeY, nodeZ);
	}

	/**
	 * Sets the z-coordinate for the given layer (graph).
	 */
	public void setZ(Graph graph, double z) {
		LayerLayout25D gl = getLayout(graph);
		gl.setZ(z);
	}

	/**
	 * Returns the z-corrdinate of the geiven layer (graph).
	 */
	public double getZ(Graph graph) {
		LayerLayout25D gl = getLayout(graph);
		return gl.getZ();
	}

}