/**
 * y25 - Visualizing Graphs in 2.5 Dimensions
 * 
 * Arbeitsbereich: Paralleles Rechnen Prof. Kaufmann
 * Studienarbeit:  Werner Jainek
 * Betreuung:      Nina Lehmann
 *
 */
package y25.view;

import y.base.Edge;
import y.base.Graph;
import y.base.Node;
import y.view.EdgeRealizer;
import y.view.Graph2D;
import y.view.NodeRealizer;
import y25.base.GraphCursor;
import y25.base.GraphTools;
import y25.base.LayerEdge;
import y25.layout.LayeredLayoutGraph;
import y25.view.realizer.PolyLineEdgeRealizer25D;
import y25.view.realizer.PolyLineLayerEdgeRealizer25D;
import y25.view.realizer.Realizer25DFactory;
import y25.view.realizer.ShapeNodeRealizer25D;
import y25.view.realizer.TransparentLayerRealizer25D;

/**
 * Represents a <code>LayeredLayoutGraph</code> with additional 2.5D support.
 * Instances of this class can be displayed within a {@link Graph25DView}. The
 * layout and visual representation of nodes, edges, layer edges and layers are
 * encapsulated in so called realizer objects, which implement the respective
 * Layout25D interfaces defined in the <code>y25.layout</code> package. Every
 * node has a <code>NodeRealizer25D</code>, every edge an
 * <code>EdgeRealizer25D</code>, every layer edge a
 * <code>LayerEdgeRealizer25D</code> and every layer (graph) a
 * <code>LayerRealizer25D</code> associated with it.
 * <p>
 * All geometric or visual changes hat are performed on this Graph25D result in
 * an attribute or type change of an associated realizer. If nodes, edges, layer
 * edges or layers (graphs) get newly created in this Graph25D then a default
 * node, edge, layer edge or layer realizer 25D will be automatically associated
 * with these elements. To change the default appearance of these elements, one
 * can assign different default realizers by using the methods
 * <code>setDefault[Node|Edge|LayerEdge|Layer]Realizer</code>.
 * <p>
 * In order to be able to assign node and edge realizers to nodes and edges
 * inside different layers, every graph this class class stores internally is of
 * type <code>Graph2D</code>, so when iterating over this Graph25D's layers,
 * casting to <code>Graph2D</code> is always allowed.
 * <p>
 * There is one subtle implementation detail worth noting: Since compatibility
 * with the yFiles package was one of the major goals in designing this package,
 * every <code>Graph2D</code> that is stored inside this <code>Graph25D</code>
 * can be displayed and modified in a separate <code>Graph2DView</code>. To
 * allow for this usage, node and edge realizers stored inside the
 * <code>Graph2D</code>s have to be of type <code>NodeRealizer</code> and
 * <code>EdgeRealizer</code> respectively. The same realizers, however, are
 * used when displaying the <code>Graph2D</code>s inside the
 * <code>Graph25DView</code>. For this to be possible, every realizer stored
 * inside a <code>Graph2D</code> needs to extend <code>NodeRealizer</code>
 * and implement the <code>NodeRealizer25D</code> interface. (<code>ShapeNodeRealizer25D</code>
 * and <code>PolyLineEdgeRealizer25D</code> are examples for this).
 * 
 * @author Werner Jainek
 */

public class Graph25D extends LayeredLayoutGraph {

	// we need to store XXXRealizer instead of XXXRealizer2D, since
	// XXXRealizer2D is not a descendant of XXXRealizer! However, all
	// NodeRealizers this Graph25D will ever store are (extends NodeRealizer
	// implements NodeRealizer25D), so a cast is always possible !

	/**
	 * The default node realizer used for newly created nodes.
	 */
	NodeRealizer defaultNodeRealizer;

	/**
	 * The default edge realizer used for newly created edges.
	 */
	EdgeRealizer defaultEdgeRealizer;

	/**
	 * The default layer edge realizer used for newly created layer edges.
	 */
	LayerEdgeRealizer25D defaultLayerEdgeRealizer;

	/**
	 * The default layer realizer used for newly created layers (graphs).
	 */
	LayerRealizer25D defaultLayerRealizer;

	/**
	 * Creates a new <code>Graph25D</code>.
	 */
	public Graph25D() {
		graphType = Graph2D.class;
		defaultNodeRealizer = new ShapeNodeRealizer25D();
		defaultEdgeRealizer = new PolyLineEdgeRealizer25D();
		defaultLayerRealizer = new TransparentLayerRealizer25D();
		defaultLayerEdgeRealizer = new PolyLineLayerEdgeRealizer25D();
	}

	/**
	 * Resets the <code>Graph25D</code> to its initial state.
	 */
	public void clear() {
		super.clear();
		defaultNodeRealizer = new ShapeNodeRealizer25D();
		defaultEdgeRealizer = new PolyLineEdgeRealizer25D();
		defaultLayerRealizer = new TransparentLayerRealizer25D();
		defaultLayerEdgeRealizer = new PolyLineLayerEdgeRealizer25D();
	}

	/**
	 * Removes all layer edges present in this <code>Graph25D</code>.
	 */
	public void clearLayerEdges() {
	}

	// GRAPHS --------------------------------------------

	/**
	 * Creates a new <code>Graph2D</code> and returns it. The new graph will
	 * be located at the end of the graph list inside this <code>Graph2D</code>.
	 * <p>
	 * The returned graph automatically assigns new and unique IDs to nodes and
	 * edges as soon as they are created. After creation, changing these
	 * automatically assigned IDs is accomplished by calling the LayeredGraph
	 * methods <code>setNodeID</code>, <code>setParentID</code> and
	 * <code>setEdgeID</code>.
	 * <p>
	 * The created graph is automatically assigned the default layer realizer.
	 * Also, newly created nodes and edges inside the <code>Graph2D</code>
	 * will automatically receive the default node and edge realizers associated
	 * with this <code>Graph25D</code>.
	 * <p>
	 * The z-position of the returned graph is set to <code>0</code>.
	 * 
	 * @return The newly created graph.
	 */
	public Graph createGraph() {

		return createGraph(AFTER, null);
	}

	/**
	 * Creates a new <code>Graph2D</code> at a specified location inside the
	 * graph list and returns it.
	 * <p>
	 * The returned graph automatically assigns new and unique IDs to nodes and
	 * edges as soon as they are created. After creation, changing these
	 * automatically assigned IDs is accomplished by calling the LayeredGraph
	 * methods <code>setNodeID</code>, <code>setParentID</code> and
	 * <code>setEdgeID</code>.
	 * <p>
	 * The created graph is automatically assigned the default layer realizer.
	 * Also, newly created nodes and edges inside the <code>Graph2D</code>
	 * will automatically receive the default node and edge realizers associated
	 * with this <code>Graph25D</code>.
	 * 
	 * @param where
	 *            Can be either <code>LayeredGraph.BEFORE</code> or
	 *            <code>LayeredGraph.AFTER</code>. In the fromer case, the
	 *            new graph is inserted into the list right before the passed
	 *            reference graph, in the latter case it is inserted after the
	 *            passed reference graph.
	 * @param reference
	 *            A graph that is already in the graph list of this Graph25D. If
	 *            <code>null</code> is passed, the newly created graph is
	 *            inserted at the very beginning of the list if
	 *            <code>where</code> is <code>LayeredGraph.BEFORE</code> or
	 *            at the very end of the graph list if <code>where</code> is
	 *            <code>LayeredGraph.AFTER</code>.
	 * 
	 * @return The newly created graph.
	 */
	public Graph createGraph(int where, Graph reference) {

		Graph2D g = (Graph2D) super.createGraph(where, reference);

		// ----- GRAPH 25D PART -----
		g.setDefaultNodeRealizer(defaultNodeRealizer);
		g.setDefaultEdgeRealizer(defaultEdgeRealizer);
		setRealizer(g, defaultLayerRealizer);

		// ----- LAYERED LAYOUT GRAPH PART -----
		recalculateBoundingBox();

		return g;
	}

	/**
	 * Creates a new <code>Graph2D</code> by copying the passed
	 * <code>Graph2D</code>. The passed graph has to be part of this
	 * <code>Graph25D</code>. The new graph's location inside the graph list
	 * will be right after the original graph.
	 * <p>
	 * The created graph is assigned a copy of the layer realizer of the
	 * original graph. Newly created nodes and edges inside the
	 * <code>Graph2D</code> will automatically receive the default node and
	 * edge realizers associated with the original graph.
	 * <p>
	 * The new graph is then returned. It has exactly the same nodes and edges
	 * as the original graph. All node, edge and parent IDs are also identical
	 * to the original. The returned graph automatically assigns new and unique
	 * IDs to newlu created nodes and edges. After creation, changing these
	 * automatically assigned IDs is accomplished by calling the LayeredGraph
	 * methods <code>setNodeID</code>, <code>setParentID</code> and
	 * <code>setEdgeID</code>.
	 * 
	 * @param original
	 *            the graph that will be copied.
	 * @return the newly created graph that is a duplicate of the original.
	 */
	public Graph duplicateGraph(Graph original) {

		// ----- LAYERED GRAPH PART -----
		// TODO: At this moment, the entire function from LayeredGraph is copied
		// here, because we need to call GraphTools.DuplicateGraph2D instead of
		// DuplicateGraph. Making this more intelligent would reduce the
		// redundancy !

		// create information about which datamaps to copy from the original
		// graph
		GraphTools.GraphDataInfo dataInfo[] = {
				new GraphTools.GraphDataInfo(NODE_ID_DPKEY,
						GraphTools.GraphDataInfo.NODEMAP,
						GraphTools.GraphDataInfo.INTEGER),
				new GraphTools.GraphDataInfo(PARENT_ID_DPKEY,
						GraphTools.GraphDataInfo.NODEMAP,
						GraphTools.GraphDataInfo.INTEGER),
				new GraphTools.GraphDataInfo(EDGE_ID_DPKEY,
						GraphTools.GraphDataInfo.EDGEMAP,
						GraphTools.GraphDataInfo.INTEGER) };

		// create the realizer factory used for creating new realizers
		Realizer25DFactory realizerFactory = new Realizer25DFactory(
				defaultNodeRealizer, defaultEdgeRealizer);

		// create a copy of the given graph
		Graph2D copy = GraphTools.DuplicateGraph2D((Graph2D) original,
				dataInfo, realizerFactory);

		// register graph listener that attaches appropriate id's to newly
		// created nodes/edges
		copy.addGraphListener(idAssigner);

		// insert copy into graph list
		graphList.insertAfter(copy, graphList.findCell(original));

		// ----- GRAPH 25D PART -----
		copy.setDefaultNodeRealizer(((Graph2D) original)
				.getDefaultNodeRealizer());
		copy.setDefaultEdgeRealizer(((Graph2D) original)
				.getDefaultEdgeRealizer());
		setRealizer(copy, getRealizer(original));

		// ----- LAYERED LAYOUT GRAPH PART -----
		recalculateBoundingBox();

		return copy;
	}

	/**
	 * Creates a new <code>Graph2D</code> by copying the passed
	 * <code>Graph2D</code>. The difference to the
	 * <code>duplicateGraph</code> method is, that the passed graph is not
	 * part of this <code>Graph25D</code>. If you already have code that
	 * generates a <code>Graph2D</code> and you want to use that graph inside
	 * this <code>Graph25D</code>, this is the method you have to use. (It is
	 * an error to pass a graph that is already inside this
	 * <code>Graph25D</code> to this method.)
	 * <p>
	 * The new graph will be located at the end of the graph list inside this
	 * <code>Graph25D</code>. It has exactly the same nodes and edges as the
	 * original graph.
	 * <p>
	 * Regarding node, edge and parent IDs, there are two possibilities:
	 * <ol>
	 * <li> If the graph you pass has no <code>DataProvider</code>s for the
	 * keys <code>LayeredGraph.NODE_ID_DPKEY</code>,
	 * <code>LayeredGraph.PARENT_ID_DPKEY</code> and
	 * <code>LayeredGraph.EDGE_ID_DPKEY</code>, new IDs will be created for
	 * every node and edge in the copied graph. The parent IDs will all be set
	 * to <code>0</code>. </li>
	 * <li> If you want to specify node, edge and parent IDs, you have to
	 * register <code>DataProvider</code>s for the following keys:
	 * <code>LayeredGraph.NODE_ID_DPKEY</code>,
	 * <code>LayeredGraph.PARENT_ID_DPKEY</code>,
	 * <code>LayeredGraph.EDGE_ID_DPKEY</code>. Then you can assign node,
	 * edge and parent IDs as you like. These IDs will then be copied to the new
	 * graph that is created. Note that you have to perform these steps
	 * <em>before</em> you call this method. </li>
	 * </ol>
	 * The new graph automatically assigns new and unique IDs to newly created
	 * nodes and edges. After creation, changing these automatically assigned
	 * IDs is accomplished by calling the LayeredGraph methods
	 * <code>setNodeID</code>, <code>setParentID</code> and
	 * <code>setEdgeID</code>.
	 * 
	 * @param graph
	 *            the graph that will be copied.
	 * @return the newly created graph that is a duplicate of the passed graph.
	 */
	public Graph importGraph(Graph graph) {

		// ----- LAYERED GRAPH PART -----
		// TODO: At this moment, the entire function from LayeredGraph is copied
		// here, because we need to call GraphTools.DuplicateGraph2D instead of
		// DuplicateGraph. Making this more intelligent would reduce the
		// redundancy !

		// create the realizer factory used for creating new realizers
		Realizer25DFactory realizerFactory = new Realizer25DFactory(
				defaultNodeRealizer, defaultEdgeRealizer);

		// copy the graph without the structure information
		Graph2D copy = GraphTools.DuplicateGraph2D((Graph2D) graph, null,
				realizerFactory);

		// copy or create the necessary structure data
		copyCreateStructureData(graph, copy);

		// register graph listener that attaches appropriate id's to newly
		// created nodes/edges
		copy.addGraphListener(idAssigner);

		// insert copy into graph list
		graphList.addLast(copy);

		// ----- GRAPH 25D PART -----
		copy.setDefaultNodeRealizer(realizerFactory
				.copyNodeRealizer(((Graph2D) graph).getDefaultNodeRealizer()));
		copy.setDefaultEdgeRealizer(realizerFactory
				.copyEdgeRealizer(((Graph2D) graph).getDefaultEdgeRealizer()));
		setRealizer(copy, defaultLayerRealizer);

		// ----- LAYERED LAYOUT GRAPH PART -----
		recalculateBoundingBox();

		return copy;
	}

	/**
	 * Removes the passed <code>Graph2D</code> from this <code>Graph25D</code>.
	 * <p>
	 * See {@link y25.base.LayeredGraph#removeGraph LayeredGraph.removeGraph}
	 * and
	 * {@link y25.layout.LayeredLayoutGraph#removeGraph LayeredLayoutGraph.removeGraph}
	 * for a description of what exactly this method does.
	 * 
	 * @param graph
	 *            The graph that will be removed.
	 * 
	 */
	public void removeGraph(Graph graph) {

		// ----- GRAPH2D PART -----
		// / nothing to do

		// ----- LAYERED GRAPH + LAYERED LAYOUT GRAPH PART -----
		super.removeGraph(graph);
	}

	// LAYER EDGES ---------------------------------------

	/**
	 * Creates a new layer edge from the source to the target node. Both nodes
	 * must be inside <code>Graph2D</code>s that belong to this
	 * <code>Graph25D</code>.
	 * <p>
	 * The created layer edge is automatically assigned the default layer edge
	 * realizer.
	 * 
	 * @param source
	 *            The source node for the layer edge.
	 * @param target
	 *            The target node for the layer edge.
	 * @return The newly created layer edge.
	 */
	public LayerEdge createLayerEdge(Node source, Node target) {

		// ----- LAYERED GRAPH + LAYERED LAYOUT GRAPH PART -----
		LayerEdge le = super.createLayerEdge(source, target);

		// ----- GRAPH25D PART -----
		LayerEdgeRealizer25D ler = defaultLayerEdgeRealizer.createCopy();
		ler.bindLayerEdge(le);
		layerEdgeLayouts.set(le, ler);

		return le;
	}

	/**
	 * Removes the passed layer edge from this Graph25D.
	 * <p>
	 * See
	 * {@link y25.base.LayeredGraph#removeLayerEdge LayeredGraph.removeLayerEdge}
	 * for a description of what exactly this method does.
	 * 
	 * @param layerEdge
	 *            The layer edge that will be removed.
	 */
	public void removeLayerEdge(LayerEdge layerEdge) {

		// ----- GRAPH25D PART -----
		// nothing to do

		// ----- LAYERED GRAPH + LAYERED LAYOUT GRAPH PART -----
		super.removeLayerEdge(layerEdge);
	}

	// SETTING / GETTING REALIZERS
	// ------------------------------------------------

	/**
	 * Sets the passed node's realizer to the passed node realizer. A copy of
	 * the realizer will be created and stored internally.
	 * <p>
	 * Note that the passed NodeRealizer has to implement the NodeRealizer25D
	 * interface. If this is not the case, an attempt is made to create an
	 * analog 25D version of the passed realizer via an instance of
	 * <code>Realizer25DFactory</code>. If this fails too, the default layer
	 * realizer of this <code>Graph25D</code> is used.
	 */
	public void setRealizer(Node node, NodeRealizer nr) {
		Graph2D g2d = (Graph2D) node.getGraph();

		g2d.setRealizer(node, new Realizer25DFactory(defaultNodeRealizer,
				defaultEdgeRealizer).copyNodeRealizer(nr));
	}

	/**
	 * Sets the passed edge's realizer to the passed edge realizer. A copy of
	 * the realizer will be created and stored internally.
	 * <p>
	 * Note that the passed EdgeRealizer has to implement the EdgeRealizer25D
	 * interface. If this is not the case, an attempt is made to create an
	 * analog 25D version of the passed realizer via an instance of
	 * <code>Realizer25DFactory</code>. If this fails too, the default layer
	 * realizer of this <code>Graph25D</code> is used.
	 */
	public void setRealizer(Edge edge, EdgeRealizer er) {
		Graph2D g2d = (Graph2D) edge.getGraph();

		g2d.setRealizer(edge, new Realizer25DFactory(defaultNodeRealizer,
				defaultEdgeRealizer).copyEdgeRealizer(er));

	}

	/**
	 * Sets the passed layer edge's realizer to the passed layer edge realizer.
	 * A copy of the realizer will be created and stored internally.
	 */
	public void setRealizer(LayerEdge layerEdge, LayerEdgeRealizer25D ler) {
		LayerEdgeRealizer25D copy = ler.createCopy();
		copy.bindLayerEdge(layerEdge);
		layerEdgeLayouts.set(layerEdge, copy);
	}

	/**
	 * Sets the passed graph's realizer to the passed layer realizer. A copy of
	 * the realizer will be created and stored internally.
	 */
	public void setRealizer(Graph graph, LayerRealizer25D lr) {
		layerLayouts.set(graph, lr.createCopy());
	}

	/**
	 * Returns the <code>NodeRealizer</code> that is associated with the
	 * passed node.
	 * <p>
	 * All <code>NodeRealizer</code>s stored inside this
	 * <code>Graph25D</code> (or inside the node's <code>Graph2D</code> for
	 * that matter) implement the <code>NodeRealizer25D</code> interface, so
	 * casting the result to that type is always safe.
	 */
	public NodeRealizer getRealizer(Node node) {
		Graph2D g2d = (Graph2D) node.getGraph();

		return g2d.getRealizer(node);
	}

	/**
	 * Returns the <code>EdgeRealizer</code> that is associated with the
	 * passed edge.
	 * <p>
	 * All <code>EdgeRealizer</code>s stored inside this
	 * <code>Graph25D</code> (or inside the edge's <code>Graph2D</code> for
	 * that matter) implement the <code>EdgeRealizer25D</code> interface, so
	 * casting the result to that type is always safe.
	 */
	public EdgeRealizer getRealizer(Edge edge) {
		Graph2D g2d = (Graph2D) edge.getGraph();

		return g2d.getRealizer(edge);
	}

	/**
	 * Returns the <code>LayerEdgeRealizer25D</code> that is associated with
	 * the passed layer edge.
	 */
	public LayerEdgeRealizer25D getRealizer(LayerEdge layerEdge) {
		return (LayerEdgeRealizer25D) layerEdgeLayouts.get(layerEdge);
	}

	/**
	 * Returns the <code>LayerRealizer25D</code> that is associated with the
	 * passed graph.
	 */
	public LayerRealizer25D getRealizer(Graph graph) {
		return (LayerRealizer25D) layerLayouts.get(graph);
	}

	// SETTING DEFAULT REALIZERS
	// ------------------------------------------------

	/**
	 * Sets the default <code>NodeRealizer</code> for this
	 * <code>Graph25D</code> and all its associated <code>Graph2D</code>s
	 * to the passsed realizer.
	 * <p>
	 * Note that the passed NodeRealizer has to implement the NodeRealizer25D
	 * interface. If this is not the case, an attempt is made to create an
	 * analog 25D version of the passed realizer via an instance of
	 * <code>Realizer25DFactory</code>. If this fails too, the default layer
	 * realizer of this <code>Graph25D</code> is used.
	 */
	public void setDefaultNodeRealizer25D(NodeRealizer nr) {
		defaultNodeRealizer = new Realizer25DFactory(defaultNodeRealizer,
				defaultEdgeRealizer).copyNodeRealizer(nr);
		for (GraphCursor gc = graphs(); gc.ok(); gc.next()) {
			((Graph2D) gc.graph()).setDefaultNodeRealizer(defaultNodeRealizer);
		}
	}

	/**
	 * Sets the default <code>EdgeRealizer</code> for this
	 * <code>Graph25D</code> and all its associated <code>Graph2D</code>s
	 * to the passsed realizer.
	 * <p>
	 * Note that the passed EdgeRealizer has to implement the EdgeRealizer25D
	 * interface. If this is not the case, an attempt is made to create an
	 * analog 25D version of the passed realizer via an instance of
	 * <code>Realizer25DFactory</code>. If this fails too, the default layer
	 * realizer of this <code>Graph25D</code> is used.
	 */
	public void setDefaultEdgeRealizer25D(EdgeRealizer er) {
		defaultEdgeRealizer = new Realizer25DFactory(defaultNodeRealizer,
				defaultEdgeRealizer).copyEdgeRealizer(er);
		for (GraphCursor gc = graphs(); gc.ok(); gc.next()) {
			((Graph2D) gc.graph()).setDefaultEdgeRealizer(defaultEdgeRealizer);
		}
	}

	/**
	 * Sets the default <code>LayerEdgeRealizer25D</code> for this
	 * <code>Graph25D</code> to the passsed realizer.
	 */
	public void setDefaultLayerEdgeRealizer25D(LayerEdgeRealizer25D ler) {
		defaultLayerEdgeRealizer = ler.createCopy();
	}

	/**
	 * Sets the default <code>LayerRealizer25D</code> for this
	 * <code>Graph25D</code> to the passsed realizer.
	 */
	public void setDefaultLayerRealizer25D(LayerRealizer25D lr) {
		defaultLayerRealizer = lr.createCopy();
	}

	/**
	 * Returns a copy of the default node realizer used in this
	 * <code>Graph25D</code>.
	 * <p>
	 * All <code>NodeRealizer</code>s stored inside this
	 * <code>Graph25D</code> and its associated <code>Graph2D</code>s
	 * implement the <code>NodeRealizer25D</code> interface, so casting the
	 * result to that type is always safe.
	 * <p>
	 * Note that this method doesn't work like its corresponding
	 * <code>Graph2D</code> method, since a <em>copy</em> of the default
	 * realizer is returned at not the realizer itself. This is done in order to
	 * prevent unexpected behavior. If the the goal is to change the default
	 * realizer of this <code>Graph25D</code> and all its associated
	 * <code>Graph2D</code>s by modifying a parameter of the default node
	 * realizer, the following steps have to be performed:
	 * <ul>
	 * <li>1. get a copy of the default node realizer by calling this method</li>
	 * <li>2. perform the required adjustments on the copy</li>
	 * <li>3. call <code>setDefaultNodeRealizer25D</code> and pass it the
	 * modified copy</li>
	 * </ul>
	 */
	public NodeRealizer getDefaultNodeRealizer25D() {
		return defaultNodeRealizer.createCopy();
	}

	/**
	 * Returns a copy of the default edge realizer used in this
	 * <code>Graph25D</code>.
	 * <p>
	 * All <code>EdgeRealizer</code>s stored inside this
	 * <code>Graph25D</code> and its associated <code>Graph2D</code>s
	 * implement the <code>EdgeRealizer25D</code> interface, so casting the
	 * result to that type is always safe.
	 * <p>
	 * Note that this method doesn't work like its corresponding
	 * <code>Graph2D</code> method, since a <em>copy</em> of the default
	 * realizer is returned at not the realizer itself. This is done in order to
	 * prevent unexpected behavior. If the the goal is to change the default
	 * realizer of this <code>Graph25D</code> and all its associated
	 * <code>Graph2D</code>s by modifying a parameter of the default edge
	 * realizer, the following steps have to be performed:
	 * <ul>
	 * <li>1. get a copy of the default edge realizer by calling this method</li>
	 * <li>2. perform the required adjustments on the copy</li>
	 * <li>3. call <code>setDefaultEdgeRealizer25D</code> and pass it the
	 * modified copy</li>
	 * </ul>
	 */
	public EdgeRealizer getDefaultEdgeRealizer25D() {
		return defaultEdgeRealizer.createCopy();
	}

	/**
	 * Returns a copy of the default layer edge realizer used in this
	 * <code>Graph25D</code>.
	 * <p>
	 * Note that this method doesn't necessarily work as you might expect. It is
	 * important to realize that a <em>copy</em> of the default realizer is
	 * returned. If the the goal is to change the default layer realizer of this
	 * <code>Graph25D</code> and all its associated <code>Graph2D</code>s
	 * by modifying a parameter of the default layer edge realizer, the
	 * following steps have to be performed:
	 * <ul>
	 * <li>1. get a copy of the default layer edge realizer by calling this
	 * method</li>
	 * <li>2. perform the required adjustments on the copy</li>
	 * <li>3. call <code>setDefaultLayerEdgeRealizer25D</code> and pass it
	 * the modified copy</li>
	 * </ul>
	 */
	public LayerEdgeRealizer25D getDefaultLayerEdgeRealizer25D() {
		return defaultLayerEdgeRealizer.createCopy();
	}

	/**
	 * Returns a copy of the default layer realizer used in this
	 * <code>Graph25D</code>.
	 * <p>
	 * Note that this method doesn't necessarily work as you might expect. It is
	 * important to realize that a <em>copy</em> of the default realizer is
	 * returned. If the the goal is to change the default layer realizer of this
	 * <code>Graph25D</code> and all its associated <code>Graph2D</code>s
	 * by modifying a parameter of the default layer realizer, the following
	 * steps have to be performed:
	 * <ul>
	 * <li>1. get a copy of the default layer realizer by calling this method</li>
	 * <li>2. perform the required adjustments on the copy</li>
	 * <li>3. call <code>setDefaultLayerRealizer25D</code> and pass it the
	 * modified copy</li>
	 * </ul>
	 */
	public LayerRealizer25D getDefaultLayerRealizer25D() {
		return defaultLayerRealizer.createCopy();
	}

}
