/**
 * y25 - Visualizing Graphs in 2.5 Dimensions
 * 
 * Arbeitsbereich: Paralleles Rechnen Prof. Kaufmann
 * Studienarbeit:  Werner Jainek
 * Betreuung:      Nina Lehmann
 *
 */
package y25.view.realizer;

import y.view.EdgeRealizer;
import y.view.NodeRealizer;
import y.view.PolyLineEdgeRealizer;
import y.view.ShapeNodeRealizer;
import y25.base.GraphTools;
import y25.view.EdgeRealizer25D;
import y25.view.NodeRealizer25D;

/**
 * Copies node and edge realizers and adds the "25D" extension, if necessarry.
 * <p>
 * It is important that all node and edge realizers that are stored inside the
 * <code>Graph2D</code>s of a <code>Graph25D</code> extend the
 * <code>Node/EdgeRealizer</code> classes as defined in the
 * <code>y.view</code> package and implement the
 * <code>Node/EdgeRealizer25D</code> interfaces defined in the
 * <code>y25.view</code> package. This necessary in order to be able to both
 * display the <code>Graph2D</code>s inside a <code>GraphView</code> and
 * inside a <code>Graph25DView</code> - the <code>Node/EdgeRealizer</code>
 * provides the 2D drawing mechanisms and the <code>Node/EdgeRealizer25D</code>
 * provides the OpenGl drawing mechanisms. A realizer that also implements the
 * "25D" interface is said to have the "25D" extension.
 * <p>
 * When copying realizers, this class provides a convenient way to ensure that
 * the result of the copy process is always a realizer with the "25D" extension,
 * regardless of whether or not the original realizer had this extension or not.
 * This is very useful when importing <code>Graph2D</code>s into a
 * <code>Graph25D</code> for instance.
 * <p>
 * To use this class you have to create an instance of it by passing it default
 * node and edge realizers that have the "25D" extension. If a node or edge
 * realizer has to be copied, {@link #copyNodeRealizer} and
 * {@link #copyEdgeRealizer} have to be called. Those methods first check if the
 * passed realizer already has the 25D extension. If so, it is just copied and
 * returned. If this is not the case, those methods check if there is a known
 * 25D version of the passed realizer. If this is the case, an instance will be
 * created and as much information as possible will be copied from the passed
 * realizer. Finally, if no 25D version of the passed realizer is found, a copy
 * of the default realizer that was spacified when creating the factory will be
 * returned.
 * <p>
 * Note that as more and more realizers from the <code>y.view</code> package
 * get a <code>y25.view.realizer</code> counterpart, this class should be
 * extended to include those "25D"-realizers when searching for a match of the
 * passed realizers.
 * 
 * @author Werner Jainek
 * 
 */
public class Realizer25DFactory implements GraphTools.RealizerFactory {

	/**
	 * The default node realizer that is used as a fallback.
	 */
	NodeRealizer defaultNodeRealizer;

	/**
	 * The default edge realizer that is used as a fallback.
	 */
	EdgeRealizer defaultEdgeRealizer;

	/**
	 * Creates a new <code>Realizer25DFactory</code> with the passsed
	 * realizers as the default node and edge realizers.
	 * 
	 * Note that the passed realizers have to implement the "25D" extension.
	 */
	public Realizer25DFactory(NodeRealizer defNR, EdgeRealizer defER) {
		defaultNodeRealizer = defNR.createCopy();
		defaultEdgeRealizer = defER.createCopy();
	}

	/**
	 * Returns a copy of the default node realizer of this factory.
	 */
	public NodeRealizer createDefaultNodeRealizer() {
		return defaultNodeRealizer.createCopy();
	}

	/**
	 * Returns a copy of the default edge realizer of this factory.
	 */
	public EdgeRealizer createDefaultEdgeRealizer() {
		return defaultEdgeRealizer.createCopy();
	}

	/**
	 * Creates a "25D" version of the passed node realizer as outlined in the
	 * class description above.
	 */
	public NodeRealizer copyNodeRealizer(NodeRealizer nr) {

		// does it already have the 25D extension ?
		if (nr instanceof NodeRealizer25D) {
			return nr.createCopy();
		}

		// is it of a known class ?
		else if (nr instanceof ShapeNodeRealizer) {
			return new ShapeNodeRealizer25D(nr);
		}

		// if all fails, just return a copy of the default node realizer
		return defaultNodeRealizer.createCopy(nr);
	}

	/**
	 * Creates a "25D" version of the passed edge realizer as outlined in the
	 * class description above.
	 */
	public EdgeRealizer copyEdgeRealizer(EdgeRealizer er) {

		// does it already have the 25D extension ?
		if (er instanceof EdgeRealizer25D) {
			return er.createCopy();
		}

		// is it of a known class ?
		else if (er instanceof PolyLineEdgeRealizer) {
			return new PolyLineEdgeRealizer25D(er);
		}

		// if all fails, just return a copy of the default node realizer
		return defaultEdgeRealizer.createCopy(er);
	}

}
