/**
 * y25 - Visualizing Graphs in 2.5 Dimensions
 * 
 * Arbeitsbereich: Paralleles Rechnen Prof. Kaufmann
 * Studienarbeit:  Werner Jainek
 * Betreuung:      Nina Lehmann
 *
 */
package y25.view.realizer;

import java.awt.Color;

import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;

import y.view.NodeLabel;
import y.view.NodeRealizer;
import y.view.ShapeNodeRealizer;
import y25.graphics.BoundingBox;
import y25.graphics.Camera;
import y25.graphics.Point3D;
import y25.graphics.Transform;
import y25.graphics.Vector3D;
import y25.view.NodeRealizer25D;

import com.sun.opengl.util.GLUT;

/**
 * Extends the <code>ShapeNodeRealizer</code> with 25D information.
 * <p>
 * This realizer aims to provide the same functionality as the
 * {@link y.view.ShapeNodeRealizer y.view.ShapeNodeRealizer}, except that it
 * can also be used inside a <code>Graph25D</code> and therefore rendered into
 * a <code>Graph25DView</code>.
 * 
 * @author Werner Jainek
 */
public class ShapeNodeRealizer25D extends ShapeNodeRealizer implements
		NodeRealizer25D {

	/**
	 * The default depth of a node.
	 */
	static double defaultDepth = 10.0;

	/**
	 * The depth of this realizer.
	 */
	protected double depth;

	/**
	 * Creates a new <code>ShapeNodeRealizer25D</code> with the default depth.
	 */
	public ShapeNodeRealizer25D() {
		depth = defaultDepth;
	}

	/**
	 * Creates a new <code>ShapeNodeRealizer25D</code> that copies as many
	 * attributes of the passed realizer as possible.
	 */
	public ShapeNodeRealizer25D(NodeRealizer nr) {
		super(nr);

		if (nr instanceof NodeRealizer25D) {
			depth = ((NodeRealizer25D) nr).getDepth();

			if (nr instanceof ShapeNodeRealizer25D) {
				// member variable copies go here !
			}
		} else {
			depth = defaultDepth;
		}
	}

	/**
	 * @see y.view.NodeRealizer#createCopy(y.view.NodeRealizer)
	 */
	public NodeRealizer createCopy(NodeRealizer nr) {
		return new ShapeNodeRealizer25D(nr);
	}

	/**
	 * @see y25.layout.NodeLayout25D#getDepth()
	 */
	public double getDepth() {
		return depth;
	}

	/**
	 * @see y25.layout.NodeLayout25D#setDepth(double)
	 */
	public void setDepth(double depth) {
		this.depth = Math.abs(depth);
	}

	/**
	 * @see y25.layout.NodeLayout25D#setSize(double, double, double)
	 */
	public void setSize(double width, double height, double depth) {
		setSize(width, height);
		setDepth(depth);
	}

	/**
	 * @see y25.layout.NodeLayout25D#getRelativeBoundingBox()
	 */
	public BoundingBox getRelativeBoundingBox() {
		BoundingBox bb = new BoundingBox();
		bb.includeX((float) getX());
		bb.includeX((float) (getX() + getWidth()));
		bb.includeY((float) getY());
		bb.includeY((float) (getY() + getHeight()));
		bb.includeZ((float) (depth * 0.5));
		bb.includeZ((float) (-depth * 0.5));
		return bb;
	}

	/**
	 * @see y25.view.NodeRealizer25D#paintGL(javax.media.opengl.GLAutoDrawable,
	 *      y25.graphics.Camera)
	 */
	public void paintGL(GLAutoDrawable drawable, Camera camera) {
		GL gl = drawable.getGL();
		GLUT glut = new GLUT();

		gl.glDisable(GL.GL_COLOR_MATERIAL);

		// specify the material
		Color nodeColor = getFillColor();
		float mat_none[] = { 0.0f, 0.0f, 0.0f, 1.0f };
		float mat_diffuse[] = { ((float) nodeColor.getRed()) / 255.0f,
				((float) nodeColor.getGreen()) / 255.0f,
				((float) nodeColor.getBlue()) / 255.0f, 1.0f };
		float mat_diffuse_selected[] = { Math.max(mat_diffuse[0] - 0.3f, 0.0f),
				Math.max(mat_diffuse[1] - 0.3f, 0.0f),
				Math.max(mat_diffuse[2] - 0.3f, 0.0f), 1.0f };
		float mat_noshininess[] = { 0.0f };
		if (!isSelected()) {
			gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT_AND_DIFFUSE,
					mat_diffuse, 0);
		} else {
			gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT_AND_DIFFUSE,
					mat_diffuse_selected, 0);
		}
		gl.glMaterialfv(GL.GL_FRONT, GL.GL_SPECULAR, mat_none, 0);
		gl.glMaterialfv(GL.GL_FRONT, GL.GL_SHININESS, mat_noshininess, 0);
		gl.glMaterialfv(GL.GL_FRONT, GL.GL_EMISSION, mat_none, 0);

		// determine the shape type
		byte shapeType = getShapeType();
		switch (shapeType) {

		case ELLIPSE:
			gl.glPushMatrix();
			gl.glTranslatef((float) getCenterX(), (float) getCenterY(), 0.0f);
			gl.glScalef((float) getWidth(), (float) getHeight(),
					(float) getDepth());
			glut.glutSolidSphere(0.5f, 10, 10);
			gl.glPopMatrix();
			break;

		default:
		case RECT:
			gl.glPushMatrix();
			gl.glTranslatef((float) getCenterX(), (float) getCenterY(), 0.0f);
			gl.glScalef((float) getWidth(), (float) getHeight(),
					(float) getDepth());
			glut.glutSolidCube(1.0f);
			gl.glPopMatrix();
			break;
		}

		// paint the label (only one supported)
		// TODO: add more sophisticated label support
		if (labelCount() > 0) {

			// find out where to place the label
			Point3D labelPos = new Point3D((float) getCenterX(),
					(float) getCenterY(), 0);
			Vector3D toCam = camera.toCamera(drawable, labelPos).normalize();
			float maxRadius = (float) Math.max(Math
					.max(getWidth(), getHeight()), getDepth());
			labelPos.translate(toCam.scale(maxRadius + 0.001f));
			NodeLabel label = getLabel();
			LabelPainter.paintGL(drawable, camera, label, labelPos);
		}

		gl.glEnable(GL.GL_COLOR_MATERIAL);

	}

	/**
	 * @see y25.view.NodeRealizer25D#intersect(y25.graphics.Point3D,
	 *      y25.graphics.Point3D, float)
	 */
	public Point3D intersect(Point3D source, Point3D target, float nodeZ) {

		Point3D intersection = new Point3D();

		Vector3D dir = new Vector3D(source, target);
		float dist = dir.length();
		dir.scale(1.0f / dist);

		Point3D center = new Point3D((float) getCenterX(),
				(float) getCenterY(), nodeZ);
		float hitDistance;
		float tmin, tmax;

		switch (getShapeType()) {

		default:
		case RECT:

			Vector3D delta = new Vector3D((float) (getWidth() / 2.0),
					(float) (getHeight() / 2.0), (float) (getDepth() / 2.0));
			Point3D pmin = center.getTranslated(delta.getReversed());
			Point3D pmax = center.getTranslated(delta);

			// determine min and max hits in each dimension
			// x direction
			float tx1 = (pmin.x - source.x) / dir.x;
			float tx2 = (pmax.x - source.x) / dir.x;
			float txmin = Math.min(tx1, tx2);
			float txmax = Math.max(tx1, tx2);
			// y direction
			float ty1 = (pmin.y - source.y) / dir.y;
			float ty2 = (pmax.y - source.y) / dir.y;
			float tymin = Math.min(ty1, ty2);
			float tymax = Math.max(ty1, ty2);
			// z direction
			float tz1 = (pmin.z - source.z) / dir.z;
			float tz2 = (pmax.z - source.z) / dir.z;
			float tzmin = Math.min(tz1, tz2);
			float tzmax = Math.max(tz1, tz2);

			// determine tmin and tmax
			tmin = Math.max(txmin, Math.max(tymin, tzmin));
			tmax = Math.min(txmax, Math.min(tymax, tzmax));

			// intersection at all ?
			if (tmin > tmax)
				return null;

			// which one is closest and valid ?
			if (tmin > 0 && tmin <= dist) {
				hitDistance = tmin;
			} // outside
			else if (tmax > 0 && tmax <= dist) {
				hitDistance = tmax;
			} // inside
			else
				return null;

			// compute the intersection
			intersection.set(source);
			intersection.translate(dir.getScaled(hitDistance));
			return intersection;

		case ELLIPSE:
			Transform object2World = Transform.scale((float) getWidth(),
					(float) getHeight(), (float) getDepth());
			Transform world2Object = object2World.getInverted();
			Vector3D s = new Vector3D(center, source);
			Vector3D t = new Vector3D(center, target);
			world2Object.apply(s);
			world2Object.apply(t);

			// do hit-testing on a normal sphere
			float radius = 0.5f;
			Vector3D d = new Vector3D(t).subtract(s).normalize(); // could use
			// world2Object.apply(dir).normalize()
			// instead

			float a = s.dot(d.getReversed());
			float b = d.dot(d);
			float c = s.dot(s) - (radius * radius);

			float x = a * a - b * c;

			// do we have only one or no intersections
			if (x <= 0)
				return null;

			// we have two intersections
			x = (float) Math.sqrt(x);
			x /= b;
			tmax = a + x;
			tmin = a - x;

			// determine closest intersection
			if (tmin > 0 && tmin <= dist) {
				hitDistance = tmin;
			} // outside
			else if (tmax > 0 && tmax <= dist) {
				hitDistance = tmax;
			} // inside
			else
				return null;

			// compute the intersection
			intersection.set(s);
			intersection.translate(d.getScaled(hitDistance));

			// go back to world coordinates
			object2World.apply(intersection);
			intersection.translate(center.getVector3D());

			return intersection;
		}

	}

}
